import inspect


__all__ = (
    'CogMeta',
    'Cog',
)

from tg_botting._types import _BaseCommand


class CogMeta(type):
    """A metaclass for defining a cog.
    
    Note that you should probably not use this directly. It is exposed
    purely for documentation purposes along with making custom metaclasses to intermix
    with other metaclasses such as the :class:`abc.ABCMeta` metaclass.
    
    For example, to create an abstract cog mixin class, the following would be done.
    
    .. code-block:: python3
    
        import abc
        class CogABCMeta(CogMeta, abc.ABCMeta):
            pass
        class SomeMixin(metaclass=abc.ABCMeta):
            pass
        class SomeCogMixin(SomeMixin, Cog, metaclass=CogABCMeta):
            pass
            
    .. note::
    
        When passing an attribute of a metaclass that is documented below, note
        that you must pass it as a keyword-only argument to the class creation
        like the following example:
        
        .. code-block:: python3
        
            class MyCog(Cog, name='My Cog'):
                pass
                
    Attributes
    -----------
    name: :class:`str`
        The cog name. By default, it is the name of the class with no modification.
    command_attrs: :class:`dict`
        A list of attributes to apply to every command inside this cog. The dictionary
        is passed into the :class:`Command` (or its subclass) options at ``__init__``.
        If you specify attributes inside the command attribute in the class, it will
        override the one specified inside this attribute. For example:
        
        .. code-block:: python3
        
            class MyCog(Cog, command_attrs=dict(hidden=True)):
                @command()
                async def foo(self, ctx):
                    pass # hidden -> True
                    
                @command(hidden=False)
                async def bar(self, ctx):
                    pass # hidden -> False
                    
    """
    def __new__(mcs, *args, **kwargs):
        name, bases, attrs = args
        attrs['__cog_name__'] = kwargs.pop('name', name)
        attrs['__cog_settings__'] = command_attrs = kwargs.pop('command_attrs', {})

        commands = {}
        listeners = {}
        no_bot_cog = 'Commands or listeners must not start with cog_ or bot_ (in method {0.__name__}.{1})'

        new_cls = super().__new__(mcs, name, bases, attrs, **kwargs)
        for base in reversed(new_cls.__mro__):
            for elem, value in base.__dict__.items():
                if elem in commands:
                    del commands[elem]
                if elem in listeners:
                    del listeners[elem]

                is_static_method = isinstance(value, staticmethod)
                if is_static_method:
                    value = value.__func__
                if isinstance(value, _BaseCommand):
                    if is_static_method:
                        raise TypeError('Command in method {0}.{1!r} must not be staticmethod.'.format(base, elem))
                    if elem.startswith(('cog_', 'bot_')):
                        raise TypeError(no_bot_cog.format(base, elem))
                    commands[elem] = value
                elif inspect.iscoroutinefunction(value):
                    try:
                        is_listener = getattr(value, '__cog_listener__')
                    except AttributeError:
                        continue
                    else:
                        if elem.startswith(('cog_', 'bot_')):
                            raise TypeError(no_bot_cog.format(base, elem))
                        listeners[elem] = value

        new_cls.__cog_commands__ = list(commands.values())  # this will be copied in Cog.__new__

        listeners_as_list = []
        for listener in listeners.values():
            for listener_name in listener.__cog_listener_names__:
                listeners_as_list.append((listener_name, listener.__name__))

        new_cls.__cog_listeners__ = listeners_as_list
        return new_cls

    def __init__(cls, *args, **kwargs):
        super().__init__(*args)

    @classmethod
    def qualified_name(mcs):
        return mcs.__cog_name__


def _cog_special_method(func):
    func.__cog_special_method__ = None
    return func


class Cog(metaclass=CogMeta):
    """The base class that all cogs must inherit from.
    
    A cog is a collection of commands, listeners, and optional state to
    help group commands together. More information on them can be found on
    the :ref:`vk_api_cogs` page.
    
    When inheriting from this class, the options shown in :class:`CogMeta`
    are equally valid here.
    """
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        cmd_attrs = cls.__cog_settings__

        self.__cog_commands__ = tuple(c._update_copy(cmd_attrs) for c in cls.__cog_commands__)

        lookup = {
            cmd.qualified_name: cmd
            for cmd in self.__cog_commands__
        }

        for command in self.__cog_commands__:
            setattr(self, command.callback.__name__, command)
            parent = command.parent
            if parent is not None:
                parent = lookup[parent.qualified_name]

                removed = parent.remove_command(command.name)
                parent.add_command(command)

        return self

    def get_commands(self):
        return [c for c in self.__cog_commands__ if c.parent is None]

    @property
    def qualified_name(self):
        """:class:`str`: Returns the cog's specified name, not the class name."""
        return self.__cog_name__

    @property
    def description(self):
        """:class:`str`: Returns the cog's description, typically the cleaned docstring."""
        try:
            return self.__cog_cleaned_doc__
        except AttributeError:
            self.__cog_cleaned_doc__ = cleaned = inspect.getdoc(self)
            return cleaned

    def walk_commands(self):
        """An iterator that recursively walks through this cog's commands and subcommands."""
        from vk_botting.commands import GroupMixin
        for command in self.__cog_commands__:
            if command.parent is None:
                yield command
                if isinstance(command, GroupMixin):
                    yield from command.walk_commands()

    def get_listeners(self):
        """Returns a :class:`list` of (name, function) listener pairs that are defined in this cog."""
        return [(name, getattr(self, method_name)) for name, method_name in self.__cog_listeners__]

    @classmethod
    def _get_overridden_method(cls, method):
        """Return None if the method is not overridden. Otherwise returns the overridden method."""
        return getattr(method.__func__, '__cog_special_method__', method)

    @classmethod
    def listener(cls, name=None):
        """A decorator that marks a function as a listener.
        
        This is the cog equivalent of :meth:`.Bot.listen`.
        
        Parameters
        ------------
        name: :class:`str`
            The name of the event being listened to. If not provided, it
            defaults to the function's name.
            
        Raises
        --------
        TypeError
            The function is not a coroutine function or a string was not passed as
            the name.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError('Cog.listener expected str but received {0.__class__.__name__!r} instead.'.format(name))

        def decorator(func):
            actual = func
            if isinstance(actual, staticmethod):
                actual = actual.__func__
            if not inspect.iscoroutinefunction(actual):
                raise TypeError('Listener function must be a coroutine function.')
            actual.__cog_listener__ = True
            to_assign = name or actual.__name__
            try:
                actual.__cog_listener_names__.append(to_assign)
            except AttributeError:
                actual.__cog_listener_names__ = [to_assign]
            return func

        return decorator

    @_cog_special_method
    def cog_unload(self):
        """A special method that is called when the cog gets removed.
        
        This function **cannot** be a coroutine. It must be a regular
        function.
        
        Subclasses must replace this if they want special unloading behaviour.
        """
        pass

    @_cog_special_method
    def bot_check_once(self, ctx):
        """A special method that registers as a :meth:`.Bot.check_once`
        check.
        
        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def bot_check(self, ctx):
        """A special method that registers as a :meth:`.Bot.check_once`
        check.
        
        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_check(self, ctx):
        """A special method that registers as a :func:`commands.check`
        for every command and subcommand in this cog.
        
        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_command_error(self, ctx, error):
        """A special method that is called whenever an error
        is dispatched inside this cog.
        
        This is similar to :func:`.on_command_error` except only applying
        to the commands inside this cog.
        
        This function **can** be a coroutine.
        
        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context where the error happened.
        error: :class:`CommandError`
            The error that happened.
        """
        pass

    @_cog_special_method
    async def cog_before_invoke(self, ctx):
        """A special method that acts as a cog local pre-invoke hook.
        This is similar to :meth:`.Command.before_invoke`.
        
        This **must** be a coroutine.
        
        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass

    @_cog_special_method
    async def cog_after_invoke(self, ctx):
        """A special method that acts as a cog local post-invoke hook.
        This is similar to :meth:`.Command.after_invoke`.
        
        This **must** be a coroutine.
        
        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass

    def _inject(self, bot):
        cls = self.__class__

        for index, command in enumerate(self.__cog_commands__):
            command.cog = self
            if command.parent is None:
                try:
                    bot.add_command(command)
                except Exception as e:
                    for to_undo in self.__cog_commands__[:index]:
                        bot.remove_command(to_undo)
                    raise e

        if cls.bot_check is not Cog.bot_check:
            bot.add_check(self.bot_check)

        if cls.bot_check_once is not Cog.bot_check_once:
            bot.add_check(self.bot_check_once, call_once=True)

        for name, method_name in self.__cog_listeners__:
            bot.add_listener(getattr(self, method_name), name)

        return self

    def _eject(self, bot):
        cls = self.__class__

        try:
            for command in self.__cog_commands__:
                if command.parent is None:
                    bot.remove_command(command.name)

            for _, method_name in self.__cog_listeners__:
                bot.remove_listener(getattr(self, method_name))

            if cls.bot_check is not Cog.bot_check:
                bot.remove_check(self.bot_check)

            if cls.bot_check_once is not Cog.bot_check_once:
                bot.remove_check(self.bot_check_once, call_once=True)
        finally:
            self.cog_unload()



import asyncio
import enum
import sys
import textwrap
import traceback
import typing
from typing import TypeVar

import aiohttp

from tg_botting.exceptions import TGException, TGApiError, BadArgument
from tg_botting.general import convert_params
from tg_botting.message import Chat, Message, CallbackQuery
from tg_botting.objects import get_chat_member, MessageEntity, InlineKeyboardMarkup, ReplyKeyboardMarkup, \
    ReplyKeyboardRemove
from tg_botting.user import User
from tg_botting.utils import to_json, maybe_coroutine

String = TypeVar('String', bound=str)
Integer = TypeVar('Integer', bound=int)
Float = TypeVar('Float', bound=float)
Boolean = TypeVar('Boolean', bound=bool)
T = TypeVar('T')

class UserMessageFlags(enum.IntFlag):
    Unread = 1,
    Outbox = 2,
    Replied = 4,
    Important = 8,
    Chat = 16,
    Friends = 32,
    Spam = 64,
    Deleted = 128,
    Fixed = 256,
    Media = 512,
    Hidden = 65536,
    DeleteForAll = 131072,
    NotDelivered = 262144


class _ClientEventTask(asyncio.Task):
    def __init__(self, original_coro, event_name, coro, *, loop):
        super().__init__(coro, loop=loop)
        self.__event_name = event_name
        self.__original_coro = original_coro

    def __repr__(self):
        info = [
            ('state', self._state.lower()),
            ('event', self.__event_name),
            ('coro', repr(self.__original_coro)),
        ]
        if self._exception is not None:
            info.append(('exception', repr(self._exception)))
        return '<ClientEventTask {}>'.format(' '.join('%s=%s' % t for t in info))


class Client:
    def __init__(self, **kwargs):
        self.force = kwargs.get('force', False)
        self.lang = kwargs.get('lang', None)
        self.loop = asyncio.get_event_loop()
        self.group = None
        self.user = None
        self.key = None
        self.server = None
        self._listeners = {}
        timeout = aiohttp.ClientTimeout(total=100, connect=10)
        user_agent = kwargs.get('user_agent', None)
        if user_agent:
            headers = {
                'User-Agent': user_agent
            }
            self.session = kwargs.get('session', aiohttp.ClientSession(timeout=timeout, headers=headers))
        else:
            self.session = kwargs.get('session', aiohttp.ClientSession(timeout=timeout))
        self._all_events = ['message_new', 'sticker' 'message_event', 'message_reply', 'message_allow',
                            'message_deny',
                            'message_edit', 'message_typing_state', 'photo_new', 'audio_new', 'video_new',
                            'wall_reply_new', 'wall_reply_edit', 'wall_reply_delete', 'wall_reply_restore',
                            'wall_post_new', 'wall_repost', 'board_post_new', 'board_post_edit', 'board_post_restore',
                            'board_post_delete', 'photo_comment_new', 'photo_comment_edit', 'photo_comment_delete',
                            'photo_comment_restore', 'video_comment_new', 'video_comment_edit', 'video_comment_delete',
                            'video_comment_restore', 'market_comment_new', 'market_comment_edit',
                            'market_comment_delete', 'market_comment_restore', 'poll_vote_new', 'group_join',
                            'group_leave', 'group_change_settings', 'group_change_photo', 'group_officers_edit',
                            'user_block', 'user_unblock']
        self.extra_events = []
        self.token = None
        self.user_token = None
        self.event_handlers = {
            'callback_query': self.handle_callback_query,
            'message_reply': self.handle_message_reply,
            'message_edit': self.handle_message_edit,
            'sticker': self.handle_sticker_new,
            # 'message_typing_state': self.handle_message_typing_state,
            # 'message_allow': self.handle_message_allow,
            # 'message_deny': self.handle_message_allow,
            # 'photo_new': self.handle_photo_new,
            # 'photo_comment_new': self.handle_photo_comment_new,
            # 'photo_comment_edit': self.handle_photo_comment_new,
            # 'photo_comment_restore': self.handle_photo_comment_new,
            # 'photo_comment_delete': self.handle_photo_comment_delete,
            # 'audio_new': self.handle_audio_new,
            # 'video_new': self.handle_video_new,
            # 'video_comment_new': self.handle_video_comment_new,
            # 'video_comment_edit': self.handle_video_comment_new,
            # 'video_comment_restore': self.handle_video_comment_new,
            # 'video_comment_delete': self.handle_video_comment_delete,
            # 'wall_post_new': self.handle_wall_post_new,
            # 'wall_repost': self.handle_wall_post_new,
            # 'wall_reply_new': self.handle_wall_reply_new,
            # 'wall_reply_edit': self.handle_wall_reply_new,
            # 'wall_reply_restore': self.handle_wall_reply_new,
            # 'wall_reply_delete': self.handle_wall_reply_delete,
            # 'board_post_new': self.handle_board_post_new,
            # 'board_post_edit': self.handle_board_post_new,
            # 'board_post_restore': self.handle_board_post_new,
            # 'board_post_delete': self.handle_board_post_delete,
            # 'market_comment_new': self.handle_market_comment_new,
            # 'market_comment_edit': self.handle_market_comment_new,
            # 'market_comment_restore': self.handle_market_comment_new,
            # 'market_comment_delete': self.handle_market_comment_delete,
            # 'group_leave': self.handle_group_leave,
            # 'group_join': self.handle_group_join,
            # 'user_block': self.handle_user_block,
            # 'user_unblock': self.handle_user_unblock,
            # 'poll_vote_new': self.handle_poll_vote_new,
            # 'group_officers_edit': self.handle_group_officers_edit,
        }

    def Payload(self, **kwargs):
        kwargs['access_token'] = self.token
        return kwargs

    def UserPayload(self, **kwargs):
        if not self.user_token:
            raise TGException('User Token not attached. Use Bot.attach_user_token to attach it.')
        kwargs['access_token'] = self.user_token
        return kwargs

    class botCommandException(Exception):
        pass

    def wait_for(self, event, *, check=None, timeout=None):
        """|coro|

        Waits for an event to be dispatched.

        This could be used to wait for a user to reply to a message or to edit a message in a self-containedway.

        The ``timeout`` parameter is passed onto :func:`asyncio.wait_for`. By default,
        it does not timeout. Note that this does propagate the
        :exc:`asyncio.TimeoutError` for you in case of timeout and is provided for
        ease of use.

        In case the event returns multiple arguments, a :class:`tuple` containing those
        arguments is returned instead. Please check the
        :ref:`documentation <vk_api_events>` for a list of events and their
        parameters.

        This function returns the **first event that meets the requirements**.

        Examples
        ---------
        Waiting for a user reply: ::

            @bot.command()
            async def greet(ctx):
                await ctx.send('Say hello!')
                def check(m):
                    return m.text == 'hello' and m.from_id == ctx.from_id
                msg = await bot.wait_for('message_new', check=check)
                await ctx.send('Hello {.from_id}!'.format(msg))

        Parameters
        ------------
        event: :class:`str`
            The event name, similar to the :ref:`event reference <vk_api_events>`,
            but without the ``on_`` prefix, to wait for.
        check: Optional[Callable[..., :class:`bool`]]
            A predicate to check what to wait for. The arguments must meet the
            parameters of the event being waited for.
        timeout: Optional[:class:`float`]
            The number of seconds to wait before timing out and raising
            :exc:`asyncio.TimeoutError`.

        Raises
        -------
        asyncio.TimeoutError
            If a timeout is provided and it was reached.

        Returns
        --------
        Any
            Returns no arguments, a single argument, or a :class:`tuple` of multiple
            arguments that mirrors the parameters passed in the
            :ref:`event reference <vk_api_events>`.
        """

        future = self.loop.create_future()
        if check is None:
            def _check(*_):
                return True

            check = _check

        ev = event.lower()
        try:
            listeners = self._listeners[ev]
        except KeyError:
            listeners = []
            self._listeners[ev] = listeners

        listeners.append((future, check))
        return asyncio.wait_for(future, timeout)

    async def general_request(self, url, post=False, **params):
        params = convert_params(params)
        for tries in range(5):
            try:
                req = self.session.post(url, data=params) if post else self.session.get(url, params=params)
                async with req as r:
                    if r.content_type == 'application/json':
                        return await r.json()
                    return await r.text()
            except Exception as e:
                print('Got exception in request: {}\nRetrying in {} seconds'.format(e, tries * 2 + 1), file=sys.stderr)
                await asyncio.sleep(tries * 2 + 1)

    async def _tg_request(self, method, post, calln=1, **kwargs):
        if calln > 10:
            raise TGApiError('TG API call failed after 10 retries')
        for param in kwargs:
            if isinstance(kwargs[param], (list, tuple)):
                kwargs[param] = ','.join(map(str, kwargs[param]))
            elif isinstance(kwargs[param], dict):
                kwargs[param] = to_json(kwargs[param])
        res = await self.general_request('https://api.telegram.org/bot{}/{}'.format(kwargs['access_token'], method),
                                         post=post,
                                         **kwargs)
        if isinstance(res, str):
            await asyncio.sleep(0.1)
            return await self._tg_request(method, post, calln + 1, **kwargs)
        error = res.get('error', None)
        if error and error.get('error_code', None) == 6:
            await asyncio.sleep(1)
            return await self._tg_request(method, post, calln + 1, **kwargs)
        elif error and error.get('error_code', None) == 10 and 'could not check access_token now' in error.get(
                'error_msg', ''):
            await asyncio.sleep(0.1)
            return await self._tg_request(method, post, calln + 1, **kwargs)
        return res

    async def tg_request(self, method, post=True, **kwargs):
        return await self._tg_request(method, post, **self.Payload(**kwargs))

    async def user_tg_request(self, method, post=True, **kwargs):
        return await self._tg_request(method, post, **self.UserPayload(**kwargs))

    async def get_me(self):
        user = await self.tg_request('getMe')
        if user.get('ok') != True:
            raise TGApiError('[{error_code}] {description}'.format(**user))
        user = user.get('result')
        return User(user)

    async def get_chat(self, chat_id):
        groups = await self.tg_request('getChat', chat_id=chat_id)
        if groups.get('ok') != True:
            raise TGApiError('[{error_code}] {description}'.format(**groups))
        chat = groups.get('result')
        return Chat(chat)

    async def get_chat_member(self, chat_id, user_id):
        user = await self.tg_request("getChatMember", chat_id=chat_id, user_id=user_id)
        if user.get('ok') != True:
            raise TGApiError('[{error_code}] {description}'.format(**user))
        response = user.get('result')
        return get_chat_member(response)

    def build_msg(self, msg):
        res = Message(msg)
        res.bot = self
        return res

    def generate_link(self, method: str):
        return f'https://api.telegram.org/bot{self.token}/{method}'

    async def longpoll(self, offset=None):
        if offset is None:
            offset = 0
        payload = {'offset': offset + 1}
        try:
            res = await self.general_request(self.generate_link("getUpdates"), **payload)
        except asyncio.TimeoutError:
            return offset, []
        if res['ok'] is False:
            offset = 0
        elif len(res['result']) != 0:
            offset = res['result'][len(res['result']) - 1]['update_id']
        updates = res['result']
        return offset, updates

    def handle_message(self, message):
        msg = self.build_msg(message)

        action = None
        if msg.sticker is not None:
            action = "sticker_new"
        elif msg.audio is not None:
            action = "audio_new"
        elif msg.video is not None:
            action = "video_new"
        elif msg.video_note is not None:
            action = "video_note_new"
        elif msg.voice is not None:
            action = "voice_new"
        elif msg.poll is not None:
            action = "poll_new"
        elif msg.left_chat_member is not None:
            action = "chat_member_left"
        elif msg.photo is not None:
            action = "photo_new"
        elif msg.new_chat_members is not None:
            action = 'new_chat_members'
        elif msg.successful_payment is not None:
            action = "successful_payment"

        if action is not None:
            return self.dispatch(action, msg)

        return self.dispatch('message_new', msg)

    def handle_callback_query(self, t, obj):
        event = CallbackQuery(obj)
        event.bot = self
        return self.dispatch(t, event)

    def handle_sticker_new(self, t, obj):
        msg = self.build_msg(obj)
        return self.dispatch(t, msg)

    def handle_message_reply(self, t, obj):
        msg = self.build_msg(obj)
        return self.dispatch(t, msg)

    def handle_message_edit(self, t, obj):
        msg = self.build_msg(obj)
        return self.dispatch(t, msg)

    def handle_custom_classes(self, t, obj, update):
        names = t.split("_")
        name = [r.title() for r in names]
        class_name = ''.join(name)

        if class_name not in globals():
            return self.dispatch('unknown', update)

        klass = globals()[class_name]
        instance = klass(obj)
        return self.dispatch(t, instance)

    def handle_update(self, update):
        update.pop('update_id')
        t = list(update.keys())[0]
        obj = update[t]
        on_t = 'on_{}'.format(t)
        used = (t in self._listeners and self._listeners[t]) or (on_t in self.extra_events and self.extra_events[on_t])
        unknown_used = ('unknown' in self._listeners and self._listeners['unknown']) or (
                'on_unknown' in self.extra_events and self.extra_events['on_unknown'])
        if t == 'message':
            return self.handle_message(obj)
        elif t in ['inline_query', 'chosen_inline_result', 'callback_query', 'shipping_query', 'pre_checkout_query',
                   'poll',
                   'poll_answer', 'my_chat_member', 'chat_member', 'chat_join_request']:
            return self.handle_custom_classes(t, obj, update)
        elif t in self.event_handlers and used:
            return self.loop.create_task(maybe_coroutine(self.event_handlers[t], t, obj))
        elif t not in self.event_handlers and unknown_used:
            return self.dispatch('unknown', update)

    def dispatch(self, event, *args, **kwargs):
        method = 'on_' + event
        listeners = self._listeners.get(event)
        if listeners:
            removed = []
            for i, (future, condition) in enumerate(listeners):
                if future.cancelled():
                    removed.append(i)
                    continue

                try:
                    result = condition(*args)
                except Exception as exc:
                    future.set_exception(exc)
                    removed.append(i)
                else:
                    if result:
                        if len(args) == 0:
                            future.set_result(None)
                        elif len(args) == 1:
                            future.set_result(args[0])
                        else:
                            future.set_result(args)
                        removed.append(i)

            if len(removed) == len(listeners):
                self._listeners.pop(event)
            else:
                for idx in reversed(removed):
                    del listeners[idx]

        try:
            coro = getattr(self, method)
        except AttributeError:
            pass
        else:
            self._schedule_event(coro, method, *args, **kwargs)

    async def on_error(self, event_method, *args, **kwargs):
        print('Ignoring exception in {}'.format(event_method), file=sys.stderr)
        traceback.print_exc()

    async def send_photo(self,
                         chat_id: typing.Union[Integer],
                         photo: typing.Union[String],
                         caption: typing.Optional[String] = None,
                         parse_mode: typing.Optional[String] = None,
                         caption_entities: typing.Optional[typing.List[MessageEntity]] = None,
                         message_thread_id: typing.Optional[Integer] = None,
                         disable_notification: typing.Optional[Boolean] = None,
                         protect_content: typing.Optional[Boolean] = None,
                         reply_to_message_id: typing.Optional[Integer] = None,
                         allow_sending_without_reply: typing.Optional[Boolean] = None,
                         reply_markup: typing.Union[InlineKeyboardMarkup,
                         ReplyKeyboardMarkup,
                         ReplyKeyboardRemove,
                         None] = None,
                         has_spoiler: typing.Optional[Boolean] = None,
                         ) -> Message:
        reply_markup = reply_markup.dict if reply_markup is not None else None
        caption_entities = [r.dict for r in caption_entities] if caption_entities else None
        data = {
            'chat_id':chat_id,
            'photo':photo,
            'caption':caption,
            'parse_mode':parse_mode,
            'caption_entities':caption_entities,
            'message_thread_id':message_thread_id,
            'disable_notification':disable_notification,
            'protect_content':protect_content,
            'reply_to_message_id':reply_to_message_id,
            'allow_sending_without_reply':allow_sending_without_reply,
            'reply_markup':reply_markup,
            'has_spoiler':has_spoiler
        }

        result = await self.tg_request('dendPhoto', True, **data)
        return result

    @staticmethod
    def prepare_file(payload, files, key, file):
        if isinstance(file, str):
            payload[key] = file
        elif file is not None:
            files[key] = file

    async def _run_event(self, coro, event_name, *args, **kwargs):
        try:
            await coro(*args, **kwargs)
        except asyncio.CancelledError:
            pass
        except Exception:
            try:
                await self.on_error(event_name, *args, **kwargs)
            except asyncio.CancelledError:
                pass

    def _schedule_event(self, coro, event_name, *args, **kwargs):
        wrapped = self._run_event(coro, event_name, *args, **kwargs)
        return _ClientEventTask(original_coro=coro, event_name=event_name, coro=wrapped, loop=self.loop)

    async def send_message(self, chat_id, text, message_thread_id=None, parse_mode=None, disable_web_page_preview=None,
                           disable_notification=None, entities=None,
                           protect_content=None, reply_to_message_id=None, allow_sending_without_reply=None,
                           reply_markup=None, **kwargs):
        as_user = kwargs.pop('as_user', False)
        if kwargs:
            print('Unknown parameters passed to send_message: {}'.format(', '.join(kwargs.keys())), file=sys.stderr)
            raise BadArgument
        if len(text) > 4096:
            w = textwrap.TextWrapper(width=4096, replace_whitespace=False)
            messages = w.wrap(text)
            for message in messages[:-1]:
                await self.send_message(chat_id, message, **kwargs)
            return await self.send_message(chat_id, messages[-1], message_thread_id=message_thread_id,
                                           parse_mode=parse_mode,
                                           entities=entities, disable_web_page_preview=disable_web_page_preview,
                                           disable_notification=disable_notification,
                                           protect_content=protect_content, reply_to_message_id=reply_to_message_id,
                                           allow_sending_without_reply=allow_sending_without_reply,
                                           reply_markup=reply_markup, **kwargs)
        params = {'chat_id': chat_id,
                  'text': text,
                  'message_thread_id': message_thread_id,
                  'parse_mode': parse_mode,
                  'entities': [r.dict for r in entities] if entities else None,
                  'disable_web_page_preview': disable_web_page_preview,
                  'disable_notification': disable_notification,
                  'protect_content': protect_content,
                  'reply_to_message_id': reply_to_message_id,
                  'allow_sending_without_reply': allow_sending_without_reply,
                  'reply_markup': reply_markup.dict if reply_markup else None,
                  }
        res = await self.tg_request('sendMessage', **params)
        # if not as_user else await self.user_vk_request(
        # 'messages.send', **params)
        if res.get('ok') != True:
            if res.get('error_code') == 9:
                await asyncio.sleep(1)
                return await self.send_message(chat_id, text, message_thread_id=message_thread_id,
                                               parse_mode=parse_mode,
                                               entities=entities, disable_web_page_preview=disable_web_page_preview,
                                               disable_notification=disable_notification,
                                               protect_content=protect_content, reply_to_message_id=reply_to_message_id,
                                               allow_sending_without_reply=allow_sending_without_reply,
                                               reply_markup=reply_markup, **kwargs)
            raise TGApiError('[{error_code}] {description}'.format(**res))
        if self.is_group and not as_user:
            params['from'] = self.group.dict
            params['message_id'] = res['result']['message_id']
            params['chat']  = res['result']['chat']
        else:
            pass
            # params['from_id'] = self.user.id
            # params['id'] = res['response']
        return self.build_msg(params)

    async def add_user_token(self, token):
        await self.attach_user_token(token)

    async def _run(self):
        self.is_group = True
        self.group = await self.get_me()
        self.dispatch('ready')
        updates = []
        offset = None
        while True:
            try:
                lp = self.loop.create_task(self.longpoll(offset))
                for update in updates:
                    self.handle_update(update)
                offset, updates = await lp
            except Exception as e:
                if self.use_stack_trace:
                    traceback.print_exc(file=sys.stderr)
                print('Ignoring exception in longpoll cycle:\n{}'.format(e), file=sys.stderr)
                offset+=1

    def run(self, token, user_id=None, user_hash=None):
        self.use_stack_trace = False
        self.token = token
        self.user_id = user_id
        self.user_hash = user_hash
        self.loop.create_task(self._run())
        self.loop.run_forever()

# class UserClient(Client):
#
#     def __init__(self, **kwargs):
#         user_agent = kwargs.get('user_agent',
#                                 'KateMobileAndroid/52.1 lite-445 (Android 4.4.2; SDK 19; x86; unknown Android SDK built for x86; en)')
#         kwargs.setdefault('user_agent', user_agent)
#         super().__init__(**kwargs)
#
#     async def build_user_msg(self, msg):
#         res = UserMessage(msg)
#         if res.attachments:
#             res.attachments = await get_user_attachments(res.attachments)
#         res.bot = self
#         return res
#
#     async def get_user_longpoll(self):
#         res = await self.vk_request('messages.getLongPollServer', group_id=self.group.id, lp_version=3)
#         error = res.get('error', None)
#         if error and error['error_code'] == 15:
#             raise LoginError('User has no access to messages API. Try generating token with vk_botting.auth methods')
#         elif error:
#             raise TGApiError('[{error_code}] {error_msg}'.format(**res['error']))
#         self.key = res['response']['key']
#         server = res['response']['server'].replace(r'\/', '/')
#         self.server = 'https://{}'.format(server)
#         ts = res['response']['ts']
#         return ts
#
#     async def longpoll(self, ts):
#         payload = {'key': self.key,
#                    'act': 'a_check',
#                    'ts': ts,
#                    'wait': '10'}
#         if not self.is_group:
#             payload['mode'] = 10
#         try:
#             res = await self.general_request(self.server, **payload)
#         except asyncio.TimeoutError:
#             return ts, []
#         if 'ts' not in res.keys() or 'failed' in res.keys():
#             ts = await self.get_user_longpoll()
#         else:
#             ts = res['ts']
#         updates = res.get('updates', [])
#         return ts, updates
#
#     async def handle_user_update(self, update):
#         t = update.pop(0)
#         if t == 4:
#             data = {
#                 'id': update.pop(0),
#                 'flags': UserMessageFlags(update.pop(0)),
#                 'peer_id': update.pop(0),
#                 'date': update.pop(0),
#                 'text': update.pop(1),
#                 'attachments': update.pop(1)
#             }
#             data['from_id'] = data['attachments'].pop('from', data['peer_id'])
#             msg = await self.build_user_msg(data)
#             return self.dispatch('message_new', msg)
#         elif 'on_unknown' in self.extra_events:
#             return self.dispatch('unknown', update)
#
#     async def _run(self):
#         # if owner_id and owner_id.__class__ is not int:
#         #     raise TypeError('Owner_id must be positive integer, not {0.__class__.__name__}'.format(owner_id))
#         # if owner_id and owner_id < 0:
#         #     raise TGApiError('Owner_id must be positive integer')
#         # user = await self.get_own_page()
#         # if isinstance(user, User):
#         self.is_group = False
#         self.group = Group({})
#         # self.user = user
#         ts = await self.get_user_longpoll()
#         self.dispatch('ready')
#         updates = []
#         while True:
#             try:
#                 lp = self.loop.create_task(self.longpoll(ts))
#                 for update in updates:
#                     self.loop.create_task(self.handle_user_update(update))
#                 ts, updates = await lp
#             except Exception as e:
#                 print('Ignoring exception in longpoll cycle:\n{}'.format(e), file=sys.stderr)
#                 ts = await self.get_user_longpoll()
#         # raise LoginError('Group token passed to user client')
#
#     def run(self, token, owner_id=None):
#         """A blocking call that abstracts away the event loop
#         initialisation from you.
#
#         .. warning::
#             This function must be the last function to call due to the fact that it
#             is blocking. That means that registration of events or anything being
#             called after this function call will not execute until it returns.
#
#         Parameters
#         ----------
#         token: :class:`str`
#             Bot token. Should be group token or user token with access to group
#         owner_id: :class:`int`
#             Should only be passed alongside user token. Owner id of group to connect to
#         """
#         self.token = token
#         self.user_token = token
#         self.loop.create_task(self._run(owner_id))
#         self.loop.run_forever()



from copy import deepcopy


class ChatPermissions:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.can_send_messages = data.get('can_send_messages') if "can_send_messages" in data else False
        self.can_send_audios = data.get('can_send_audios') if "can_send_audios" in data else False
        self.can_send_documents = data.get('can_send_documents') if "can_send_documents" in data else False
        self.can_send_photos = data.get('can_send_photos') if "can_send_photos" in data else False
        self.can_send_videos = data.get('can_send_videos') if "can_send_videos" in data else False
        self.can_send_video_notes = data.get('can_send_video_notes') if "can_send_video_notes" in data else False
        self.can_send_voice_notes = data.get('can_send_voice_notes') if "can_send_voice_notes" in data else False
        self.can_send_polls = data.get('can_send_polls') if "can_send_polls" in data else False
        self.can_send_other_messages = data.get(
            'can_send_other_messages') if "can_send_other_messages" in data else False
        self.can_add_web_page_previews = data.get(
            'can_add_web_page_previews') if "can_add_web_page_previews" in data else False
        self.can_change_info = data.get('can_change_info') if "can_change_info" in data else False
        self.can_invite_users = data.get('can_invite_users') if "can_invite_users" in data else False
        self.can_pin_messages = data.get('can_pin_messages') if "can_pin_messages" in data else False
        self.can_manage_topics = data.get('can_manage_topics') if "can_manage_topics" in data else False

    @classmethod
    def create(cls, can_send_messages=False, can_send_audios=False,
               can_send_documents=False, can_send_photos=False,
               can_send_videos=False, can_send_video_notes=False,
               can_send_voice_notes=False, can_send_polls=False,
               can_send_other_messages=False, can_add_web_page_previews=False,
               can_change_info=False, can_invite_users=False,
               can_pin_messages=False, can_manage_topics=False):
        data = {
            "can_send_messages": can_send_messages,
            "can_send_audios": can_send_audios,
            "can_send_documents": can_send_documents,
            "can_send_photos": can_send_photos,
            "can_send_videos": can_send_videos,
            "can_send_video_notes": can_send_video_notes,
            "can_send_voice_notes": can_send_voice_notes,
            "can_send_polls": can_send_polls,
            "can_send_other_messages": can_send_other_messages,
            "can_add_web_page_previews": can_add_web_page_previews,
            "can_change_info": can_change_info,
            "can_invite_users": can_invite_users,
            "can_pin_messages": can_pin_messages,
            "can_manage_topics": can_manage_topics
        }
        return cls(data)



import ctypes
import json
from copy import deepcopy
from datetime import datetime
from random import randint

from tg_botting.abstract import Messageable
from tg_botting.exceptions import TGApiError
from tg_botting.objects import Animation, Audio, Document, PhotoSize, Sticker, Video, VideoNote, Voice, Contact, \
    Dice, MessageEntity, Game, Poll, Venue, Location, MessageAutoDeleteTimerChanged, Invoice, SuccessfulPayment, \
    UserShared, ChatShared, WriteAccessAllowed, PassportData, ProximityAlertTriggered, ForumTopicReopened, \
    ForumTopicClosed, ForumTopicEdited, ForumTopicCreated, GeneralForumTopicHidden, GeneralForumTopicUnhidden, \
    VideoChatScheduled, VideoChatStarted, VideoChatEnded, VideoChatParticipantsInvited, WebAppData, \
    InlineKeyboardMarkup, ChatPhoto, ChatLocation, ChatInviteLink, get_params_from_class, get_chat_member
from tg_botting.permissions import ChatPermissions
from tg_botting.user import User


class ChatMemberUpdated:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.chat = Chat(data.get('chat'))
        self.from_ = User(data.get('from'))
        self.date = data.get('date')
        self.old_chat_member = get_chat_member(data.get('old_chat_member'))
        self.new_chat_member = get_chat_member(data.get('new_chat_member'))
        self.invite_link = ChatInviteLink(data.get('invite_link')) if data.get('invite_link',
                                                                               None) is not None else None
        self.via_chat_folder_invite_link = data.get('via_chat_folder_invite_link', None)


class ChatJoinRequest:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.chat = Chat(data.get('chat'))
        self.from_ = User(data.get('from'))
        self.user_chat_id = data.get('user_chat_id')
        self.date = data.get('date')
        self.bio = data.get('bio', None)
        self.invite_link = ChatInviteLink(data.get('invite_link')) if data.get('invite_link',
                                                                               None) is not None else None


class Chat:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.type = data.get('type')
        self.title = data.get('title') if "title" in data else ""
        self.username = data.get('username') if "username" in data else None
        self.first_name = data.get('first_name') if "first_name" in data else None
        self.last_name = data.get('last_name') if "last_name" in data else None
        self.is_forum = data.get('is_forum') if "is_forum" in data else False
        self.photo = ChatPhoto(data.get('photo')) if "photo" in data else None
        self.active_usernames = [r for r in data.get('active_usernames')] if "active_usernames" in data else []
        self.emoji_status_custom_emoji_id = data.get(
            "emoji_status_custom_emoji_id") if 'emoji_status_custom_emoji_id' in data else None
        self.bio = data.get("bio") if 'bio' in data else None
        self.has_private_forwards = data.get("has_private_forwards") if 'has_private_forwards' in data else False
        self.has_restricted_voice_and_video_messages = data.get(
            "has_restricted_voice_and_video_messages") if 'has_restricted_voice_and_video_messages' in data else False
        self.join_to_send_messages = data.get("join_to_send_messages") if 'join_to_send_messages' in data else False
        self.join_by_request = data.get("join_by_request") if 'join_by_request' in data else False
        self.description = data.get("description") if 'description' in data else None
        self.invite_link = data.get("invite_link") if 'invite_link' in data else None
        self.pinned_message = Message(data.get("pinned_message")) if 'pinned_message' in data else None
        self.permissions = ChatPermissions(data.get("permissions")) if 'permissions' in data else None
        self.slow_mode_delay = data.get("slow_mode_delay") if 'slow_mode_delay' in data else None
        self.message_auto_delete_time = data.get(
            "message_auto_delete_time") if 'message_auto_delete_time' in data else None
        self.has_aggressive_anti_spam_enabled = data.get(
            "has_aggressive_anti_spam_enabled") if 'has_aggressive_anti_spam_enabled' in data else False
        self.has_hidden_members = data.get("has_hidden_members") if 'has_hidden_members' in data else False
        self.has_protected_content = data.get("has_protected_content") if 'has_protected_content' in data else False
        self.sticker_set_name = data.get("sticker_set_name") if 'sticker_set_name' in data else None
        self.can_set_sticker_set = data.get("can_set_sticker_set") if 'can_set_sticker_set' in data else False
        self.linked_chat_id = data.get("linked_chat_id") if 'linked_chat_id' in data else None
        self.location = ChatLocation(data.get("location")) if 'location' in data else None

    @property
    def dict(self):
        data = get_params_from_class(self.__dict__, "original_data")
        return data


class CallbackQuery:
    __slots__ = ('bot', 'id', 'from_', 'message', 'inline_message_id', 'chat_instance', 'data', 'game_short_name')

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_ = data.get('from')
        self.message = Message(data.get('message'))
        self.inline_message_id = data.get('inline_message_id')
        self.chat_instance = data.get('chat_instance')
        self.data = data.get('data')
        self.game_short_name = data.get('game_short_name')

    async def _answer(self, text="", show_alert=True, url=None, cache_time=None):
        data = {
            'callback_query_id': self.id,
            'text': text,
            'show_alert': show_alert
        }
        if url is not None:
            data["url"] = url
        if cache_time is not None:
            data["cache_time"] = cache_time
        res = await self.bot.tg_request('answerCallbackQuery', data)
        if 'error' in res.keys():
            raise TGApiError('[{error_code}] {error_msg}'.format(**res['error']))
        return res

    async def blank_answer(self):
        return await self._answer("", False)


class Message(Messageable):

    async def _get_conversation(self):
        return self.chat.id

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.message_id = data.get('message_id')
        self.message_thread_id = data.get('message_thread_id') if "message_thread_id" in data else -1
        self.user = User(data.get("from")) if "from" in data else None
        self.sender_chat = Chat(data.get("sender_chat")) if "sender_chat" in data else None
        self.date = datetime.fromtimestamp(data.get('date', 86400))
        self.chat = Chat(data.get('chat'))
        self.forward_from_chat = Chat(data.get('forward_from_chat')) if 'forward_from_chat' in data else None
        self.message_thread_id = data.get('message_thread_id') if 'message_thread_id' in data else None
        self.forward_signature = data.get('forward_signature') if 'forward_signature' in data else None
        self.forward_sender_name = data.get('forward_sender_name') if 'forward_sender_name' in data else None
        self.forward_date = datetime.fromtimestamp(data.get('forward_date')) if 'forward_date' in data else None
        self.is_topic_message = data.get('is_topic_message') if 'is_topic_message' in data else False
        self.is_automatic_forward = data.get('is_automatic_forward') if 'is_automatic_forward' in data else False
        self.reply_to_message = Message(data.get('reply_to_message')) if 'reply_to_message' in data else None
        self.via_bot = User(data.get('via_bot')) if 'via_bot' in data else None
        self.edit_date = datetime.fromtimestamp(data.get('edit_date')) if 'edit_date' in data else None
        self.has_protected_content = data.get('has_protected_content') if 'has_protected_content' in data else False
        self.media_group_id = data.get('media_group_id') if 'media_group_id' in data else None
        self.author_signature = data.get('author_signature') if 'author_signature' in data else None
        self.text = data.get('text') if 'text' in data and data.get('text', None) is not None else None
        self.entities = [MessageEntity(r) for r in data.get('entities')] if 'entities' in data and data.get('entities',
                                                                                                            None) is not None else []
        self.animation = Animation(data.get('animation')) if 'animation' in data else None
        self.audio = Audio(data.get('audio')) if 'audio' in data else None
        self.document = Document(data.get('document')) if 'document' in data else None
        self.photo = [PhotoSize(r) for r in data.get('photo')] if 'photo' in data else None
        self.sticker = Sticker(data.get('sticker')) if 'sticker' in data else None
        self.video = Video(data.get('video')) if 'video' in data else None
        self.video_note = VideoNote(data.get('video_note')) if 'video_note' in data else None
        self.voice = Voice(data.get('voice')) if 'voice' in data else None
        self.caption = data.get('caption') if 'caption' in data else None
        self.caption_entities = [MessageEntity(r) for r in
                                 data.get('caption_entities')] if 'caption_entities' in data else None
        self.has_media_spoiler = data.get('has_media_spoiler') if 'has_media_spoiler' in data else False
        self.contact = Contact(data.get('contact')) if 'contact' in data else None
        self.dice = Dice(data.get('dice')) if 'dice' in data else None
        self.game = Game(data.get('game')) if 'game' in data else None
        self.poll = Poll(data.get('poll')) if 'poll' in data else None
        self.venue = Venue(data.get('venue')) if 'venue' in data else None
        self.location = Location(data.get('location')) if 'location' in data else None
        self.new_chat_members = [User(r) for r in data.get('new_chat_members')] if 'new_chat_members' in data else None
        self.left_chat_member = User(data.get('left_chat_member')) if 'left_chat_member' in data else None
        self.left_chat_participant = User(
            data.get('left_chat_participant')) if 'left_chat_participant' in data else None
        self.new_chat_title = data.get('new_chat_title') if 'new_chat_title' in data else None
        self.new_chat_photo = [PhotoSize(r) for r in data.get('new_chat_photo')] if 'new_chat_photo' in data else None
        self.delete_chat_photo = data.get('delete_chat_photo') if 'delete_chat_photo' in data else None
        self.group_chat_created = data.get('group_chat_created') if 'group_chat_created' in data else None
        self.supergroup_chat_created = data.get(
            'supergroup_chat_created') if 'supergroup_chat_created' in data else None
        self.channel_chat_created = data.get('channel_chat_created') if 'channel_chat_created' in data else None
        self.message_auto_delete_timer_changed = MessageAutoDeleteTimerChanged(
            data.get('message_auto_delete_timer_changed')) if 'message_auto_delete_timer_changed' in data else None
        self.migrate_to_chat_id = data.get('migrate_to_chat_id') if 'migrate_to_chat_id' in data else None
        self.migrate_from_chat_id = data.get('migrate_from_chat_id') if 'migrate_from_chat_id' in data else None
        self.pinned_message = Message(data.get('pinned_message')) if 'pinned_message' in data else None
        self.invoice = Invoice(data.get('invoice')) if 'invoice' in data else None
        self.successful_payment = SuccessfulPayment(
            data.get('successful_payment')) if 'successful_payment' in data else None
        self.user_shared = UserShared(data.get('user_shared')) if 'user_shared' in data else None
        self.chat_shared = ChatShared(data.get('chat_shared')) if 'chat_shared' in data else None
        self.connected_website = data.get('connected_website') if 'connected_website' in data else None
        self.write_access_allowed = WriteAccessAllowed(
            data.get('write_access_allowed')) if 'write_access_allowed' in data else None
        self.passport_data = PassportData(data.get('passport_data')) if 'passport_data' in data else None
        self.proximity_alert_triggered = ProximityAlertTriggered(
            data.get('proximity_alert_triggered')) if 'proximity_alert_triggered' in data else None
        self.forum_topic_created = ForumTopicCreated(
            data.get('forum_topic_created')) if 'forum_topic_created' in data else None
        self.forum_topic_edited = ForumTopicEdited(
            data.get('forum_topic_edited')) if 'forum_topic_edited' in data else None
        self.forum_topic_closed = ForumTopicClosed(
            data.get('forum_topic_closed')) if 'forum_topic_closed' in data else None
        self.forum_topic_reopened = ForumTopicReopened(
            data.get('forum_topic_reopened')) if 'forum_topic_reopened' in data else None
        self.general_forum_topic_hidden = GeneralForumTopicHidden(
            data.get('general_forum_topic_hidden')) if 'general_forum_topic_hidden' in data else None
        self.general_forum_topic_unhidden = GeneralForumTopicUnhidden(
            data.get('general_forum_topic_unhidden')) if 'general_forum_topic_unhidden' in data else None
        self.video_chat_scheduled = VideoChatScheduled(
            data.get('video_chat_scheduled')) if 'video_chat_scheduled' in data else None
        self.video_chat_started = VideoChatStarted(
            data.get('video_chat_started')) if 'video_chat_started' in data else None
        self.video_chat_ended = VideoChatEnded(data.get('video_chat_ended')) if 'video_chat_ended' in data else None
        self.video_chat_participants_invited = VideoChatParticipantsInvited(
            data.get('video_chat_participants_invited')) if 'video_chat_participants_invited' in data else None
        self.web_app_data = WebAppData(data.get('web_app_data')) if 'web_app_data' in data else None
        self.reply_markup = InlineKeyboardMarkup(data.get('reply_markup')) if 'reply_markup' in data and data.get(
            'reply_markup', None) is not None else None

    async def edit(self, text=None, *, reply_markup: InlineKeyboardMarkup = None, parse_mode: str = None,
                   entities: ctypes.Array = None, disable_web_page_preview: bool = None, message_id: int = None,
                   inline_message_id: int = None):
        params = {
            'chat_id': self.chat.id,
            'text': text
        }
        if reply_markup:
            params['reply_markup'] = json.dumps(reply_markup.dict)
        if parse_mode:
            params['parse_mode'] = parse_mode
        if entities:
            params['entities'] = [r.dict for r in entities]
        if disable_web_page_preview:
            params['disable_web_page_preview'] = disable_web_page_preview
        if message_id:
            params['message_id'] = message_id
        if inline_message_id:
            params['inline_message_id'] = inline_message_id
        res = await self.bot.tg_request('editMessageText', **params)
        if 'error' in res.keys():
            raise TGApiError('[{error_code}] {error_msg}'.format(**res['error']))
        return res

    async def delete(self):
        params = {
            'chat_id': self.chat.id,
            'message_id': self.message_id
        }
        res = await self.bot.tg_request('deleteMessage', **params)
        if res['ok']!=True:
            raise TGApiError('[{error_code}] {error_msg}'.format(**res))



    async def send_photo(self,
                         photo:str,
                         caption: str=None,**kwargs):
        res = await self.bot.send_photo(self.chat.id,photo,caption,*kwargs)
        if res['ok']!=True:
            raise TGApiError('[{error_code}] {error_msg}'.format(**res))
        return Message(res['result'])

    async def reply(self, text, **kwargs):
        return await self.bot.send_message(text, reply_to_message_id=self.message_id,**kwargs)

    async def get_user(self):
        author = await self.bot.get_pages(self.from_id)
        return author[0]

    async def get_author(self):
        return await self.get_user()

    async def fetch_user(self):
        return await self.get_user()

    async def fetch_author(self):
        return await self.get_user()


class UserMessage(Messageable):

    async def _get_conversation(self):
        return self.peer_id

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.date = data.get('date')
        self.flags = data.get('flags')
        self.peer_id = data.get('peer_id')
        self.from_id = data.get('from_id')
        self.text = data.get('text')
        self.attachments = data.get('attachments')
        self.important = data.get('important')
        self.payload = data.get('payload')
        self.keyboard = data.get('keyboard')

    async def edit(self, message=None, *, attachment=None, keep_forward_messages='true', keep_snippets='true'):
        params = {'peer_id': self.peer_id, 'message': message, 'attachment': attachment, 'message_id': self.id,
                  'keep_forward_messages': keep_forward_messages, 'keep_snippets': keep_snippets}
        res = await self.bot.vk_request('messages.edit', **params)
        if 'error' in res.keys():
            raise TGApiError('[{error_code}] {error_msg}'.format(**res['error']))
        return res

    async def reply(self, message=None, *, attachment=None, sticker_id=None, keyboard=None):
        peer_id = await self._get_conversation()
        params = {'random_id': randint(-2 ** 63, 2 ** 63 - 1), 'peer_id': peer_id, 'message': message,
                  'attachment': attachment,
                  'reply_to': self.id, 'sticker_id': sticker_id, 'keyboard': keyboard}
        res = await self.bot.vk_request('messages.send', **params)
        if 'error' in res.keys():
            raise TGApiError('[{error_code}] {error_msg}'.format(**res['error']))
        params['id'] = res['response']
        return self.bot.build_msg(params)

    async def get_user(self):
        user = await self.bot.get_pages(self.from_id)
        if user:
            return user[0]
        return None

    async def get_author(self):
        return await self.get_user()

    async def fetch_user(self):
        return await self.get_user()

    async def fetch_author(self):
        return await self.get_user()



import abc

from tg_botting.context_managers import Typing


class Messageable(metaclass=abc.ABCMeta):
    __slots__ = ('bot',)

    @abc.abstractmethod
    async def _get_conversation(self):
        raise NotImplementedError

    async def send(self, text, **kwargs):
        peer_id = await self._get_conversation()
        return await self.bot.send_message(peer_id, text, **kwargs)

    async def trigger_typing(self):
        peer_id = await self._get_conversation()
        res = await self.bot.vk_request('messages.setActivity', group_id=self.bot.group.id, type='typing',
                                        peer_id=peer_id)
        return res

    def typing(self):
        return Typing(self)



from tg_botting.abstract import Messageable
from tg_botting.utils import get_params_from_class,get_params_from_func


class User(Messageable):
    async def _get_conversation(self):
        return self.id

    def __init__(self,data):
        self.original_data = data
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.is_bot = data.get('is_bot')
        self.first_name = data.get('first_name')
        self.last_name = data.get('last_name') if "last_name" in data else ""
        self.username = data.get('username') if "username" in data else None
        self.language_code = data.get('language_code') if "language_code" in data else None
        self.is_premium = data.get('is_premium') if "is_premium" in data else False
        self.added_to_attachment_menu = data.get('added_to_attachment_menu') if "added_to_attachment_menu" in data else False
        self.can_join_groups = data.get('can_join_groups') if "can_join_groups" in data else False
        self.can_read_all_group_messages = data.get('can_read_all_group_messages') if "can_read_all_group_messages" in data else False
        self.supports_inline_queries = data.get('supports_inline_queries') if "supports_inline_queries" in data else False


    @property
    def mention(self,style=None):
        if style is None:
            return "@{}".format(self.username)
        elif style=="mardown":
            return "[{} {}](tg://user?id={})".format(self.first_name,self.last_name,self.id)
        elif style=="html":
            return "<a href=\"tg://user?id={}\">{} {}</a>".format(self.first_name,self.last_name,self.id)
        else:
            return "@{}".format(self.username) if self.username != None else self.mention("html")

    @property
    def dict(self):
        data = get_params_from_class(self.__dict__, "original_data")
        return data


from tg_botting import Command

def check(predicate):

    def decorator(func):
        if isinstance(func, Command):
            func.checks.append(predicate)
        else:
            if not hasattr(func, '__commands_checks__'):
                func.__commands_checks__ = []

            func.__commands_checks__.append(predicate)

        return func
    return decorator


def in_user_list(*ids):
    ids = list(map(int, ids))

    def predicate(ctx):
        if ctx.message.from_id in ids:
            return True

    return check(predicate)



import json
from copy import deepcopy
from datetime import datetime

from tg_botting.user import User
from tg_botting.utils import get_params_from_func, get_params_from_class


def get_chat_member(response):
    chatMember = ChatMember(response)
    if chatMember.status == 'creator':
        return ChatMemberOwner(response)
    elif chatMember.status == 'administrator':
        return ChatMemberAdministrator(response)
    elif chatMember.status == 'member':
        return ChatMemberMember(response)
    elif chatMember.status == 'restricted':
        return ChatMemberRestricted(response)
    elif chatMember.status == 'left':
        return ChatMemberLeft(response)
    elif chatMember.status == 'kicked':
        return ChatMemberBanned(response)
    return chatMember


class ChatPhoto:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.small_file_id = data.get('small_file_id')
        self.small_file_unique_id = data.get('small_file_unique_id')
        self.big_file_id = data.get('big_file_id')
        self.big_file_unique_id = data.get('big_file_unique_id')


class Location:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.longitude = data.get('longitude')
        self.latitude = data.get('latitude')
        self.horizontal_accuracy = data.get('horizontal_accuracy') if 'horizontal_accuracy' in data else 0
        self.live_period = data.get('live_period') if 'live_period' in data else 0
        self.heading = data.get('heading') if 'heading' in data else 0
        self.proximity_alert_radius = data.get('proximity_alert_radius') if 'proximity_alert_radius' in data else 0


class ChatLocation:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.location = Location(data.get("location"))
        self.address = data.get("address")


class File:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self.__unpack(data)

    def __unpack(self, data):
        self.file_id = data.get('file_id')
        self.file_unique_id = data.get('file_unique_id')
        self.width = data.get('width') if 'widht' in data else None
        self.height = data.get('height') if 'height' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None
        self.file_path = data.get('file_path') if 'file_path' in data else None


class PhotoSize(File):

    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.file_size = data.get('file_size') if 'file_size' in data else None


class Animation(File):

    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.duration = data.get('duration')
        self.thumbnail = PhotoSize(data.get('thumbnail')) if 'thumbnail' in data else None
        self.file_name = data.get('file_name') if 'file_name' in data else None
        self.mime_type = data.get('mime_type') if 'mime_type' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None


class Audio(File):

    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.duration = data.get('duration')
        self.performer = data.get('performer') if 'performer' in data else None
        self.title = data.get('title') if 'title' in data else None
        self.file_name = data.get('file_name') if 'file_name' in data else None
        self.mime_type = data.get('mime_type') if 'mime_type' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None
        self.thumbnail = PhotoSize(data.get('thumbnail')) if 'thumbnail' in data else None


class Document(File):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.thumbnail = PhotoSize(data.get('thumbnail')) if 'thumbnail' in data else None
        self.file_name = data.get('file_name') if 'file_name' in data else None
        self.mime_type = data.get('mime_type') if 'mime_type' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None


class MaskPosition:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.point = data.get('point')
        self.x_shift = data.get('x_shift')
        self.y_shift = data.get('y_shift')
        self.scale = data.get('scale')


class Sticker(File):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.type = data.get('type')
        self.is_animated = data.get('is_animated')
        self.is_video = data.get('is_video')
        self.thumbnail = PhotoSize(data.get('thumbnail')) if 'thumbnail' in data else None
        self.emoji = data.get('emoji') if 'emoji' in data else None
        self.set_name = data.get('set_name') if 'set_name' in data else None
        self.premium_animation = File(data.get('premium_animation')) if 'premium_animation' in data else None
        self.mask_position = MaskPosition(data.get('mask_position')) if 'mask_position' in data else None
        self.custom_emoji_id = data.get('custom_emoji_id') if 'custom_emoji_id' in data else None
        self.needs_repainting = data.get('needs_repainting') if 'needs_repainting' in data else False
        self.file_size = data.get('file_size') if 'file_size' in data else None


class Video(File):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.duration = data.get('duration')
        self.thumbnail = PhotoSize(data.get('thumbnail')) if 'thumbnail' in data else None
        self.file_name = data.get('file_name') if 'file_name' in data else None
        self.mime_type = data.get('mime_type') if 'mime_type' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None


class VideoNote(File):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.length = data.get('length')
        self.duration = data.get('duration')
        self.thumbnail = PhotoSize(data.get('thumbnail')) if 'thumbnail' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None


class Voice(File):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.duration = data.get('duration')
        self.mime_type = data.get('mime_type') if 'mime_type' in data else None
        self.file_size = data.get('file_size') if 'file_size' in data else None


class Contact:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.phone_number = data.get('phone_number')
        self.first_name = data.get('first_name')
        self.last_name = data.get('last_name') if 'last_name' in data else None
        self.user_id = data.get('user_id') if 'user_id' in data else None
        self.vcard = data.get('vcard	') if 'vcard' in data else None


class Dice:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.emoji = data.get('emoji')
        self.value = data.get('value')


class MessageEntity:

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.type = data.get('type')
        self.offset = data.get('offset')
        self.length = data.get('length')
        self.url = data.get('url') if 'url' in data else None
        self.user = User(data.get('user')) if 'user' in data else None
        self.language = data.get('language') if 'language' in data else None
        self.custom_emoji_id = data.get('custom_emoji_id') if 'custom_emoji_id' in data else None

    @classmethod
    def create(cls, type, offset, lenght, url=None, user=None, language=None, custom_emoji_id=None):
        data = get_params_from_func(locals(), "data")
        return cls(data)

    @property
    def dict(self):
        data = get_params_from_class(self.__dict__, "original_data")
        return data


class Game:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.title = data.get('title')
        self.description = data.get('description')
        self.photo = [PhotoSize(r) for r in data.get('photo')]
        self.text = data.get('text') if 'text' in data else None
        self.text_entities = [MessageEntity(r) for r in data.get('text_entities')] if 'text_entities' in data else None
        self.animation = Animation(data.get('animation')) if 'animation' in data else None


class PollOption:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.text = data.get('text')
        self.voter_count = data.get('voter_count')


class Poll:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.question = data.get('question')
        self.options = [PollOption(r) for r in data.get('options')]
        self.total_voter_count = data.get('total_voter_count')
        self.is_closed = data.get('is_closed')
        self.is_anonymous = data.get('is_anonymous')
        self.type = data.get('type')
        self.allows_multiple_answers = data.get('allows_multiple_answers')
        self.correct_option_id = data.get('correct_option_id') if 'correct_option_id' in data else None
        self.explanation = data.get('explanation') if 'explanation' in data else None
        self.explanation_entities = [MessageEntity(r) for r in
                                     data.get('explanation_entities')] if 'explanation_entities' in data else None
        self.open_period = data.get('open_period') if 'open_period' in data else None
        self.close_date = data.get('close_date') if 'close_date' in data else None


class Venue:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.location = Location(data.get('location'))
        self.title = data.get('title')
        self.address = data.get('address')
        self.foursquare_id = data.get('foursquare_id') if 'foursquare_id' in data else None
        self.foursquare_type = data.get('foursquare_type') if 'foursquare_type' in data else None
        self.google_place_id = data.get('google_place_id') if 'google_place_id' in data else None
        self.google_place_type = data.get('google_place_type') if 'google_place_type' in data else None


class MessageAutoDeleteTimerChanged:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.message_auto_delete_time = data.get('message_auto_delete_time')


class Invoice:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.title = data.get('title')
        self.description = data.get('description')
        self.start_parameter = data.get('start_parameter')
        self.currency = data.get('currency')
        self.total_amount = data.get('total_amount')


class ShippingAddress:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.country_code = data.get('country_code')
        self.state = data.get('state')
        self.city = data.get('city')
        self.street_line1 = data.get('street_line1')
        self.street_line2 = data.get('street_line2')
        self.post_code = data.get('post_code')


class OrderInfo:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.name = data.get('name') if 'name' in data else None
        self.phone_number = data.get('phone_number') if 'phone_number' in data else None
        self.email = data.get('email') if 'email' in data else None
        self.shipping_address = ShippingAddress(data.get('shipping_address')) if 'shipping_address' in data else None


class SuccessfulPayment:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.currency = data.get('currency')
        self.total_amount = data.get('total_amount')
        self.invoice_payload = data.get('invoice_payload')
        self.shipping_option_id = data.get('shipping_option_id') if 'shipping_option_id' in data else None
        self.order_info = OrderInfo(data.get('order_info')) if 'order_info' in data else None
        self.telegram_payment_charge_id = data.get('telegram_payment_charge_id')
        self.provider_payment_charge_id = data.get('provider_payment_charge_id')


class UserShared:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.request_id = data.get('request_id')
        self.user_id = data.get('user_id')


class ChatShared:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.request_id = data.get('request_id')
        self.chat_id = data.get('chat_id')


class WriteAccessAllowed:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.web_app_name = data.get('web_app_name') if 'web_app_name' in data else None


class PasportFile(File):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.file_size = data.get('file_size')
        self.file_date = datetime.fromtimestamp(data.get('file_date'))


class EncryptedPassportElement:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.type = data.get('type')
        self.data = data.get('data') if 'data' in data else None
        self.phone_number = data.get('phone_number') if 'phone_number' in data else None
        self.email = data.get('email') if 'email' in data else None
        self.files = [PasportFile(r) for r in data.get('files')] if 'files' in data else None
        self.front_side = PasportFile(data.get('front_side')) if 'front_side' in data else None
        self.reverse_side = PasportFile(data.get('reverse_side')) if 'reverse_side' in data else None
        self.selfie = PasportFile(data.get('selfie')) if 'selfie' in data else None
        self.translation = [PasportFile(r) for r in data.get('translation')] if 'translation' in data else None
        self.hash = data.get('hash') if 'hash' in data else None


class EncryptedCredentials:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.data = data.get('data')
        self.hash = data.get('hash')
        self.secret = data.get('secret')


class PassportData:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.data = [EncryptedPassportElement(r) for r in data.get('data')]
        self.credentials = EncryptedCredentials(data.get('credentials'))


class ProximityAlertTriggered:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.traveler = User(data.get('traveler'))
        self.watcher = User(data.get('watcher'))
        self.distance = data.get('distance')


class ForumTopicCreated:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.name = data.get('name')
        self.icon_color = data.get('icon_color')
        self.icon_custom_emoji_id = data.get('icon_custom_emoji_id') if 'icon_custom_emoji_id' in data else None


class ForumTopicEdited:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.name = data.get('name') if 'name' in data else None
        self.icon_custom_emoji_id = data.get('icon_custom_emoji_id') if 'icon_custom_emoji_id' in data else None


class ForumTopicClosed:
    def __init__(self, data):
        self.original_data = deepcopy(data)


class ForumTopicReopened:
    def __init__(self, data):
        self.original_data = deepcopy(data)


class GeneralForumTopicHidden:
    def __init__(self, data):
        self.original_data = deepcopy(data)


class GeneralForumTopicUnhidden:
    def __init__(self, data):
        self.original_data = deepcopy(data)


class VideoChatScheduled:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.start_date = data.get('start_date') if 'start_date' in data else None


class VideoChatEnded:
    def __init__(self, data):
        self.original_data = deepcopy(data)


class VideoChatStarted:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.duration = data.get('duration') if 'duration' in data else None


class VideoChatParticipantsInvited:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.users = [User(r) for r in data.get('users')] if 'users' in data else None


class WebAppData:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.data = data.get('data') if 'data' in data else None
        self.button_text = data.get('button_text') if 'button_text' in data else None


class WebAppInfo:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.url = data.get('url')

    @property
    def dict(self):
        return {'url':self.url}


class LoginUrl:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.url = data.get('url')
        self.forward_text = data.get('forward_text') if 'forward_text' in data else None
        self.bot_username = data.get('bot_username') if 'bot_username' in data else None
        self.request_write_access = data.get('request_write_access') if 'request_write_access' in data else False


class SwitchInlineQueryChosenChat:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.query = data.get('query') if 'query' in data else None
        self.allow_user_chats = data.get('allow_user_chats') if 'allow_user_chats' in data else None
        self.allow_bot_chats = data.get('allow_bot_chats') if 'allow_bot_chats' in data else None
        self.allow_group_chats = data.get('allow_group_chats') if 'allow_group_chats' in data else None
        self.allow_channel_chats = data.get('allow_channel_chats') if 'allow_channel_chats' in data else None


class CallbackGame:
    def __init__(self, data):
        self.original_data = deepcopy(data)


class InlineKeyboardButton:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.text = data.get('text')
        self.url = data.get('url') if 'url' in data else None
        self.callback_data = data.get('callback_data') if 'callback_data' in data else None
        self.web_app = WebAppInfo(data.get('web_app')) if 'web_app' in data else None
        self.login_url = LoginUrl(data.get('login_url')) if 'login_url' in data else None
        self.switch_inline_query = data.get('switch_inline_query') if 'switch_inline_query' in data else None
        self.switch_inline_query_current_chat = data.get(
            'switch_inline_query_current_chat') if 'switch_inline_query_current_chat' in data else None
        self.switch_inline_query_chosen_chat = SwitchInlineQueryChosenChat(
            data.get('switch_inline_query_chosen_chat')) if 'switch_inline_query_chosen_chat' in data else None
        self.callback_game = CallbackGame(data.get('callback_game')) if 'callback_game' in data else None
        self.pay = data.get('pay') if 'pay' in data else None

    @classmethod
    def create(cls, text, url=None, callback_data=None, web_app=None,
               login_url=None, switch_inline_query=None, switch_inline_query_current_chat=None,
               switch_inline_query_chosen_chat=None, callback_game=None, pay=None):
        d = {}
        for a, b in locals().items():
            if a == 'cls' or a == 'd':
                continue
            if b != None:
                d.update({a: b})
        return cls(d)

    @property
    def dict(self):
        d_ = {}
        for a, b in self.__dict__.items():
            if a == 'original_data':
                continue
            if b is None:
                continue
            d_.update({a: b})
        return d_


class InlineKeyboardMarkup:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.inline_keyboard = [[InlineKeyboardButton(l) for l in r]  for r in data.get('inline_keyboard')]\
            if 'inline_keyboard' in data else []

    @classmethod
    def create(cls):
        return cls({})

    def addButton(self, button: InlineKeyboardButton):
        self.inline_keyboard.append(button)

    @property
    def dict(self):
        data = {
            'inline_keyboard': [[r.dict] for r in self.inline_keyboard]
        }
        return data

class KeyboardButtonRequestUser:
    def __init__(self, request_id, user_is_bot=None,user_is_premium=None):
        self.request_id = request_id
        self.user_is_bot = user_is_bot
        self.user_is_premium = user_is_premium

    @property
    def dict(self):
        return self.__dict__

class ChatAdministratorRights:
    def __init__(self,is_anonymous=False,can_manage_chat=False,can_delete_messages=False,
                 can_manage_video_chats=False,can_restrict_members=False,can_promote_members=False,
                 can_change_info=False,can_invite_users=False,can_post_messages=False,
                 can_edit_messages=False,can_pin_messages=False,can_manage_topics=False):
        self.is_anonymous = is_anonymous
        self.can_manage_chat = can_manage_chat
        self.can_delete_messages = can_delete_messages
        self.can_manage_video_chats = can_manage_video_chats
        self.can_restrict_members =can_restrict_members
        self.can_promote_members = can_promote_members
        self.can_change_info = can_change_info
        self.can_invite_users = can_invite_users
        self.can_post_messages = can_post_messages
        self.can_edit_messages = can_edit_messages
        self.can_pin_messages = can_pin_messages
        self.can_manage_topics = can_manage_topics

    @property
    def dict(self):
        return self.__dict__

class KeyboardButtonRequestChat:
    def __init__(self, request_id, chat_is_channel,chat_is_forum=None,chat_has_username=None,chat_is_created=None,
                 user_administrator_rights:ChatAdministratorRights=None,bot_administrator_rights:ChatAdministratorRights=None,
                 bot_is_member=None):
        self.request_id = request_id
        self.user_is_bot = chat_is_channel
        self.chat_is_forum = chat_is_forum
        self.chat_has_username = chat_has_username
        self.chat_is_created = chat_is_created
        self.user_administrator_rights = user_administrator_rights.dict if user_administrator_rights else None
        self.bot_administrator_rights = bot_administrator_rights.dict if bot_administrator_rights else None
        self.bot_is_member = bot_is_member

    @property
    def dict(self):
        return self.__dict__

class KeyboardButtonPollType:
    def __init__(self,type=None):
        self.type = type

    @property
    def dict(self):
        return self.__dict__

class KeyboardButton:
    def __init__(self,text,request_user:KeyboardButtonRequestUser=None,request_chat:KeyboardButtonRequestChat=None,
                 request_contact=None,request_location=None,request_poll:KeyboardButtonPollType=None,web_app:WebAppInfo=None):
        self.text = text
        self.request_user = request_user.dict if request_user else None
        self.request_chat =request_chat.dict if request_chat else None
        self.request_contact = request_contact
        self.request_location =request_location
        self.request_poll = request_poll.dict if request_poll else None
        self.web_app = web_app.dict if web_app else None

    @property
    def dict(self):
        return self.__dict__


class ReplyKeyboardMarkup:
    def __init__(self, keyboard, is_persistent=None,resize_keyboard=None,one_time_keyboard=None,
                 input_field_placeholder=None,selective=None):
        self.keyboard = [r.dict for r in keyboard]
        self.is_persistent = is_persistent
        self.resize_keyboard = resize_keyboard
        self.one_time_keyboard = one_time_keyboard
        self.input_field_placeholder = input_field_placeholder
        self.selective =selective

    @property
    def dict(self):
        return self.__dict__

class ReplyKeyboardRemove:
    def __init__(self,remove_keyboard,selective=None):
        self.selective = selective
        self.remove_keyboard = remove_keyboard

    @property
    def dict(self):
        return self.__dict__

class ChatMember:
    def __init__(self, data):
        self.original_data = deepcopy(data)
        self.__unpack(data)

    def __unpack(self, data):
        self.status = data.get('status')
        self.user = User(data.get('user'))


class ChatMemberOwner(ChatMember):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.is_anonymous = data.get('is_anonymous')
        self.custom_title = data.get('custom_title') if 'custom_title' in data else None


class ChatMemberAdministrator(ChatMember):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.can_be_edited = data.get('can_be_edited')
        self.is_anonymous = data.get('is_anonymous')
        self.can_manage_chat = data.get('can_manage_chat')
        self.can_delete_messages = data.get('can_delete_messages')
        self.can_manage_video_chats = data.get('can_manage_video_chats')
        self.can_restrict_members = data.get('can_restrict_members')
        self.can_promote_members = data.get('can_promote_members')
        self.can_change_info = data.get('can_change_info')
        self.can_invite_users = data.get('can_invite_users')
        self.can_post_messages = data.get('can_post_messages')
        self.can_edit_messages = data.get('can_edit_messages')
        self.can_pin_messages = data.get('can_pin_messages')
        self.can_manage_topics = data.get('can_manage_topics')
        self.custom_title = data.get('custom_title') if 'custom_title' in data else None


class ChatMemberMember(ChatMember):
    def __init__(self, data):
        super().__init__(data)


class ChatMemberRestricted(ChatMember):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.is_member = data.get('is_member')
        self.can_send_messages = data.get('can_send_messages')
        self.can_send_audios = data.get('can_send_audios')
        self.can_send_documents = data.get('can_send_documents')
        self.can_send_photos = data.get('can_send_photos')
        self.can_send_videos = data.get('can_send_videos')
        self.can_send_video_notes = data.get('can_send_video_notes')
        self.can_send_voice_notes = data.get('can_send_voice_notes')
        self.can_send_polls = data.get('can_send_polls')
        self.can_send_other_messages = data.get('can_send_other_messages')
        self.can_add_web_page_previews = data.get('can_add_web_page_previews')
        self.can_change_info = data.get('can_change_info')
        self.can_invite_users = data.get('can_invite_users')
        self.can_pin_messages = data.get('can_pin_messages')
        self.can_manage_topics = data.get('can_manage_topics')
        self.until_date = data.get('until_date')


class ChatMemberLeft(ChatMember):
    def __init__(self, data):
        super().__init__(data)


class ChatMemberBanned(ChatMember):
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.until_date = data.get('until_date')


class InlineQuery:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_ = User(data.get('from'))
        self.query = data.get('query')
        self.offset = data.get('offset')
        self.chat_type = data.get('chat_type', None)
        self.location = Location(data.get('location')) if data.get('location', None) is not None else None


class ChosenInlineResult:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.result_id = data.get('result_id')
        self.from_ = User(data.get('from'))
        self.location = Location(data.get('location')) if data.get('location', None) is not None else None
        self.inline_message_id = data.get('inline_message_id', None)
        self.query = data.get('query', None)


class ShippingQuery:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id', None)
        self.invoice_payload = data.get('invoice_payload', None)
        self.shipping_address = ShippingAddress(data.get('shipping_address', None))
        self.from_ = User(data.get('from'))


class PreCheckoutQuery:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id', None)
        self.from_ = User(data.get('from'))
        self.currency = data.get('currency', None)
        self.total_amount = data.get('total_amount', None)
        self.invoice_payload = data.get('invoice_payload', None)
        self.shipping_option_id = data.get('shipping_option_id', None)
        self.order_info = OrderInfo(data.get('order_info', None)) if data.get('order_info', None) is not None else None


class PollAnswer:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.poll_id = data.get('poll_id', None)
        self.user = User(data.get('user', None))
        self.option_ids = [r for r in data.get('option_ids')] if data.get('option_ids', None) is not None else []


class ChatInviteLink:
    def __init__(self, data):
        super().__init__(data)
        self._unpack(data)

    def _unpack(self, data):
        self.invite_link = data.get('invite_link', None)
        self.creator = User(data.get('creator', None))
        self.creates_join_request = data.get('creates_join_request', None)
        self.is_primary = data.get('is_primary', None)
        self.is_revoked = data.get('is_revoked', None)
        self.name = data.get('name', None)
        self.expire_date = data.get('expire_date', None)
        self.member_limit = data.get('member_limit', None)
        self.pending_join_request_count = data.get('pending_join_request_count', None)



import asyncio
import datetime
import functools
import inspect
import sys
import typing

import tg_botting.conversions as converters
from tg_botting._types import _BaseCommand
from tg_botting.cog import Cog
from tg_botting.cooldowns import CooldownMapping, BucketType, Cooldown
from tg_botting.exceptions import CommandError, CommandInvokeError, ClientException, ConversionError, BadArgument, \
    BadUnionArgument, MissingRequiredArgument, TooManyArguments, DisabledCommand, CheckFailure, CommandOnCooldown
from tg_botting.utils import maybe_coroutine, async_all


def wrap_callback(coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        try:
            ret = await coro(*args, **kwargs)
        except CommandError:
            raise
        except asyncio.CancelledError:
            return
        except Exception as exc:
            raise CommandInvokeError(exc) from exc
        return ret

    return wrapped


def hooked_wrapped_callback(command, ctx, coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        try:
            ret = await coro(*args, **kwargs)
        except CommandError:
            ctx.command_failed = True
            raise
        except asyncio.CancelledError:
            ctx.command_failed = True
            return
        except Exception as exc:
            ctx.command_failed = True
            raise CommandInvokeError(exc) from exc
        finally:
            await command.call_after_hooks(ctx)
        return ret

    return wrapped


class GroupMixin:
    """A mixin that implements common functionality for classes that are allowed to register commands.

    Attributes
    -----------
    all_commands: :class:`dict`
        A mapping of command name to :class:`.Command` or subclass
        objects.
    case_insensitive: :class:`bool`
        Whether the commands should be case insensitive. Defaults to ``False``.
    """

    def __init__(self, *args, **kwargs):
        case_insensitive = kwargs.pop('case_insensitive', False)
        self.all_commands = _CaseInsensitiveDict() if case_insensitive else {}
        self.case_insensitive = case_insensitive
        super().__init__(*args, **kwargs)

    @property
    def commands(self):
        """Set[:class:`.Command`]: A unique set of commands without aliases that are registered."""
        return set(self.all_commands.values())

    def recursively_remove_all_commands(self):
        for command in self.all_commands.copy().values():
            if isinstance(command, GroupMixin):
                command.recursively_remove_all_commands()
            self.remove_command(command.name)

    def add_command(self, command):
        """Adds a :class:`.Command` or its subclasses into the internal list
        of commands.

        This is usually not called, instead the :meth:`~.GroupMixin.command`
        shortcut decorator are used instead.

        Parameters
        -----------
        command: :class:`Command`
            The command to add.

        Raises
        -------
        :exc:`.ClientException`
            If the command is already registered.
        TypeError
            If the command passed is not a subclass of :class:`.Command`.
        """

        if not isinstance(command, Command):
            raise TypeError('The command passed must be a subclass of Command')

        if isinstance(self, Command):
            command.parent = self

        if command.name in self.all_commands:
            raise ClientException('Command {0.name} is already registered.'.format(command))

        self.all_commands[command.name] = command
        for alias in command.aliases:
            if alias in self.all_commands:
                raise ClientException('The alias {} is already an existing command or alias.'.format(alias))
            self.all_commands[alias] = command

    def remove_command(self, name):
        """Remove a :class:`.Command` or subclasses from the internal list
        of commands.

        This could also be used as a way to remove aliases.

        Parameters
        -----------
        name: :class:`str`
            The name of the command to remove.

        Returns
        --------
        :class:`.Command` or subclass
            The command that was removed. If the name is not valid then
            `None` is returned instead.
        """
        command = self.all_commands.pop(name, None)

        if command is None:
            return None

        if name in command.aliases:
            return command

        for alias in command.aliases:
            self.all_commands.pop(alias, None)
        return command

    def walk_commands(self):
        """An iterator that recursively walks through all commands and subcommands."""
        for command in tuple(self.all_commands.values()):
            yield command
            if isinstance(command, GroupMixin):
                yield from command.walk_commands()

    def get_command(self, name):
        """Get a :class:`.Command` or subclasses from the internal list
        of commands.

        This could also be used as a way to get aliases.

        The name could be fully qualified (e.g. ``'foo bar'``) will get
        the subcommand ``bar`` of the group command ``foo``. If a
        subcommand is not found then ``None`` is returned just as usual.

        Parameters
        -----------
        name: :class:`str`
            The name of the command to get.

        Returns
        --------
        :class:`Command` or subclass
            The command that was requested. If not found, returns ``None``.
        """

        if ' ' not in name:
            return self.all_commands.get(name)

        names = name.split()
        obj = self.all_commands.get(names[0])
        if not isinstance(obj, GroupMixin):
            return obj

        for name in names[1:]:
            try:
                obj = obj.all_commands[name]
            except (AttributeError, KeyError):
                return None

        return obj

    def command(self, *args, **kwargs):
        """A shortcut decorator that invokes :func:`.command` and adds it to
        the internal command list via :meth:`~.GroupMixin.add_command`.
        """

        def decorator(func):
            kwargs.setdefault('parent', self)
            result = command(*args, **kwargs)(func)
            self.add_command(result)
            return result

        return decorator


def command(name=None, cls=None, **attrs):
    """A decorator that transforms a function into a :class:`.Command`.

    All checks added using the :func:`.check` & co. decorators are added into
    the function. There is no way to supply your own checks through this
    decorator.

    Parameters
    -----------
    name: :class:`str`
        The name to create the command with. By default this uses the
        function name unchanged.
    cls
        The class to construct with. By default this is :class:`.Command`.
        You usually do not change this.
    attrs
        Keyword arguments to pass into the construction of the class denoted
        by ``cls``.

    Raises
    -------
    TypeError
        If the function is not a coroutine or is already a command.
    """
    if cls is None:
        cls = Command

    def decorator(func):
        if isinstance(func, Command):
            raise TypeError('Callback is already a command.')
        return cls(func, name=name, **attrs)

    return decorator


class _CaseInsensitiveDict(dict):
    def __contains__(self, k):
        return super().__contains__(k.lower())

    def __delitem__(self, k):
        return super().__delitem__(k.lower())

    def __getitem__(self, k):
        return super().__getitem__(k.lower())

    def get(self, k, default=None):
        return super().get(k.lower(), default)

    def pop(self, k, default=None):
        return super().pop(k.lower(), default)

    def __setitem__(self, k, v):
        super().__setitem__(k.lower(), v)


class Command(_BaseCommand):
    r"""A class that implements the protocol for a bot text command.

    These are not created manually, instead they are created via the
    decorator or functional interface.

    Attributes
    -----------
    name: :class:`str`
        The name of the command.
    callback: :ref:`coroutine <coroutine>`
        The coroutine that is executed when the command is called.
    aliases: :class:`list`
        The list of aliases the command can be invoked under.
    enabled: :class:`bool`
        A boolean that indicates if the command is currently enabled.
        If the command is invoked while it is disabled, then
        :exc:`.DisabledCommand` is raised to the :func:`.on_command_error`
        event. Defaults to ``True``.
    parent: Optional[:class:`Command`]
        The parent command that this command belongs to. ``None`` if there
        isn't one.
    cog: Optional[:class:`Cog`]
        The cog that this command belongs to. ``None`` if there isn't one.
    checks: List[Callable[..., :class:`bool`]]
        A list of predicates that verifies if the command could be executed
        with the given :class:`.Context` as the sole parameter. If an exception
        is necessary to be thrown to signal failure, then one inherited from
        :exc:`.CommandError` should be used. Note that if the checks fail then
        :exc:`.CheckFailure` exception is raised to the :func:`.on_command_error`
        event.
    rest_is_raw: :class:`bool`
        If ``False`` and a keyword-only argument is provided then the keyword
        only argument is stripped and handled as if it was a regular argument
        that handles :exc:`.MissingRequiredArgument` and default values in a
        regular matter rather than passing the rest completely raw. If ``True``
        then the keyword-only argument will pass in the rest of the arguments
        in a completely raw matter. Defaults to ``False``.
    invoked_subcommand: Optional[:class:`Command`]
        The subcommand that was invoked, if any.
    ignore_extra: :class:`bool`
        If ``True``\, ignores extraneous strings passed to a command if all its
        requirements are met (e.g. ``?foo a b c`` when only expecting ``a``
        and ``b``). Otherwise :func:`.on_command_error` and local error handlers
        are called with :exc:`.TooManyArguments`. Defaults to ``True``.
    cooldown_after_parsing: :class:`bool`
        If ``True``\, cooldown processing is done after argument parsing,
        which calls converters. If ``False`` then cooldown processing is done
        first and then the converters are called second. Defaults to ``False``.
    """

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        self.__original_kwargs__ = kwargs.copy()
        return self

    def __init__(self, func, **kwargs):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Callback must be a coroutine')

        self.name = name = kwargs.pop('name', None) or func.__name__
        if not isinstance(name, str):
            raise TypeError('Name of a command must be a string')

        self.callback = func
        self.enabled = kwargs.pop('enabled', True)

        self.rest_is_raw = kwargs.pop('rest_is_raw', False)
        self.aliases = kwargs.pop('aliases', [])

        if not isinstance(self.aliases, (list, tuple)):
            raise TypeError("Aliases of a command must be a list of strings")

        try:
            cooldown = func.__commands_cooldown__
        except AttributeError:
            cooldown = kwargs.pop('cooldown', None)
        finally:
            self._buckets = CooldownMapping(cooldown)

        self.cooldown_after_parsing = kwargs.pop('cooldown_after_parsing', False)
        self.cog = None

        parent = kwargs.pop('parent', None)
        self.parent = parent if isinstance(parent, _BaseCommand) else None
        self._before_invoke = None
        self._after_invoke = None

        try:
            self.checks = func.__commands_checks__
            self.checks.reverse()
        except AttributeError:
            self.checks = kwargs.pop('checks', [])
        self.ignore_extra = kwargs.pop('ignore_extra', True)
        if kwargs.pop('has_spaces', False):
            print('Parameter has_spaces is deprecated as of 0.7.0. Please do not use it', file=sys.stderr)
        if kwargs:
            print('Unknown parameters passed to `{}` Command constructor: {}'.format(name, ', '.join(kwargs.keys())),
                  file=sys.stderr)

    def add_check(self, func):
        """Adds a check to the command.
        This is the non-decorator interface to :func:`.check`.

        Parameters
        -----------
        func
            The function that will be used as a check.
        """

        self.checks.append(func)

    def remove_check(self, func):
        """Removes a check from the command.
        This function is idempotent and will not raise an exception
        if the function is not in the command's checks.

        Parameters
        -----------
        func
            The function to remove from the checks.
        """

        try:
            self.checks.remove(func)
        except ValueError:
            pass

    @property
    def callback(self):
        return self._callback

    @callback.setter
    def callback(self, function):
        self._callback = function
        self.module = function.__module__

        signature = inspect.signature(function)
        self.params = signature.parameters.copy()

        for key, value in self.params.items():
            if isinstance(value.annotation, str):
                self.params[key] = value = value.replace(annotation=eval(value.annotation, function.__globals__))

    def update(self, **kwargs):
        """Updates :class:`Command` instance with updated attribute.

        This works similarly to the :func:`.command` decorator in terms
        of parameters in that they are passed to the :class:`Command` or
        subclass constructors, sans the name and callback.
        """
        self.__init__(self.callback, **dict(self.__original_kwargs__, **kwargs))

    def _ensure_assignment_on_copy(self, other):
        other._before_invoke = self._before_invoke
        other._after_invoke = self._after_invoke
        if self.checks != other.checks:
            other.checks = self.checks.copy()
        try:
            other.on_error = self.on_error
        except AttributeError:
            pass
        return other

    def copy(self):
        """Creates a copy of this command."""
        ret = self.__class__(self.callback, **self.__original_kwargs__)
        return self._ensure_assignment_on_copy(ret)

    def _update_copy(self, kwargs):
        if kwargs:
            kw = kwargs.copy()
            kw.update(self.__original_kwargs__)
            copy = self.__class__(self.callback, **kw)
            return self._ensure_assignment_on_copy(copy)
        else:
            return self.copy()

    async def dispatch_error(self, ctx, error):
        ctx.command_failed = True
        cog = self.cog
        try:
            coro = self.on_error
        except AttributeError:
            pass
        else:
            injected = wrap_callback(coro)
            if cog is not None:
                await injected(cog, ctx, error)
            else:
                await injected(ctx, error)
        try:
            if cog is not None:
                local = Cog._get_overridden_method(cog.cog_command_error)
                if local is not None:
                    wrapped = wrap_callback(local)
                    await wrapped(ctx, error)
        finally:
            ctx.bot.dispatch('command_error', ctx, error)

    async def _actual_conversion(self, ctx, converter, argument, param):
        if converter is bool:
            return converters._convert_to_bool(argument)

        try:
            module = converter.__module__
        except AttributeError:
            pass
        else:
            if module is not None and (module.startswith('vk_botting.') and not module.endswith('converter')):
                converter = getattr(converters, converter.__name__ + 'Converter')

        try:
            if inspect.isclass(converter):
                if issubclass(converter, converters.Converter):
                    instance = converter()
                    ret = await instance.convert(ctx, argument)
                    return ret
                else:
                    method = getattr(converter, 'convert', None)
                    if method is not None and inspect.ismethod(method):
                        ret = await method(ctx, argument)
                        return ret
            elif isinstance(converter, converters.Converter):
                ret = await converter.convert(ctx, argument)
                return ret
        except CommandError:
            raise
        except Exception as exc:
            raise ConversionError(converter, exc) from exc

        try:
            return converter(argument)
        except CommandError:
            raise
        except Exception as exc:
            try:
                name = converter.__name__
            except AttributeError:
                name = converter.__class__.__name__

            raise BadArgument('Converting to "{}" failed for parameter "{}".'.format(name, param.name)) from exc

    async def do_conversion(self, ctx, converter, argument, param):
        try:
            origin = converter.__origin__
        except AttributeError:
            pass
        else:
            if origin is typing.Union:
                errors = []
                _NoneType = type(None)
                for conv in converter.__args__:
                    if conv is _NoneType and param.kind != param.VAR_POSITIONAL:
                        ctx.view.undo()
                        return None if param.default is param.empty else param.default

                    try:
                        value = await self._actual_conversion(ctx, conv, argument, param)
                    except CommandError as exc:
                        errors.append(exc)
                    else:
                        return value

                raise BadUnionArgument(param, converter.__args__, errors)

        return await self._actual_conversion(ctx, converter, argument, param)

    def _get_converter(self, param):
        converter = param.annotation
        if converter is param.empty:
            if param.default is not param.empty:
                converter = str if param.default is None else type(param.default)
            else:
                converter = str
        return converter

    async def transform(self, ctx, param):
        required = param.default is param.empty
        converter = self._get_converter(param)
        consume_rest_is_special = param.kind == param.KEYWORD_ONLY and not self.rest_is_raw
        view = ctx.view
        view.skip_ws()

        if type(converter) is converters._Greedy:
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                return await self._transform_greedy_pos(ctx, param, required, converter.converter)
            elif param.kind == param.VAR_POSITIONAL:
                return await self._transform_greedy_var_pos(ctx, param, converter.converter)
            else:
                converter = converter.converter

        if view.eof:
            if param.kind == param.VAR_POSITIONAL:
                raise RuntimeError()  # break the loop
            if required:
                if self._is_typing_optional(param.annotation):
                    return None
                raise MissingRequiredArgument(param)
            return param.default

        previous = view.index
        if consume_rest_is_special:
            argument = view.read_rest().strip()
        else:
            argument = view.get_quoted_word()
        view.previous = previous

        return await self.do_conversion(ctx, converter, argument, param)

    async def _transform_greedy_pos(self, ctx, param, required, converter):
        view = ctx.view
        result = []
        while not view.eof:
            previous = view.index

            view.skip_ws()
            argument = view.get_quoted_word()
            try:
                value = await self.do_conversion(ctx, converter, argument, param)
            except CommandError:
                view.index = previous
                break
            else:
                result.append(value)

        if not result and not required:
            return param.default
        return result

    async def _transform_greedy_var_pos(self, ctx, param, converter):
        view = ctx.view
        previous = view.index
        argument = view.get_quoted_word()
        try:
            value = await self.do_conversion(ctx, converter, argument, param)
        except CommandError:
            view.index = previous
            raise RuntimeError() from None
        else:
            return value

    @property
    def clean_params(self):
        """Retrieves the parameter OrderedDict without the context or self parameters.

        Useful for inspecting signature.
        """
        result = self.params.copy()
        if self.cog is not None:
            result.popitem(last=False)

        try:
            result.popitem(last=False)
        except Exception:
            raise ValueError('Missing context parameter') from None

        return result

    @property
    def full_parent_name(self):
        """:class:`str`: Retrieves the fully qualified parent command name.

        This the base command name required to execute it. For example,
        in ``?one two three`` the parent name would be ``one two``.
        """
        entries = []
        command = self
        while command.parent is not None:
            command = command.parent
            entries.append(command.name)

        return ' '.join(reversed(entries))

    @property
    def parents(self):
        """:class:`Command`: Retrieves the parents of this command.

        If the command has no parents then it returns an empty :class:`list`.

        For example in commands ``?a b c test``, the parents are ``[c, b, a]``.
        """
        entries = []
        command = self
        while command.parent is not None:
            command = command.parent
            entries.append(command)

        return entries

    @property
    def root_parent(self):
        """Retrieves the root parent of this command.

        If the command has no parents then it returns ``None``.

        For example in commands ``?a b c test``, the root parent is ``a``.
        """
        if not self.parent:
            return None
        return self.parents[-1]

    @property
    def qualified_name(self):
        """:class:`str`: Retrieves the fully qualified command name.

        This is the full parent name with the command name as well.
        For example, in ``?one two three`` the qualified name would be
        ``one two three``.
        """
        parent = self.full_parent_name
        if parent:
            return parent + ' ' + self.name
        else:
            return self.name

    def __str__(self):
        return self.qualified_name

    async def _parse_arguments(self, ctx):
        ctx.args = [ctx] if self.cog is None else [self.cog, ctx]
        ctx.kwargs = {}
        args = ctx.args
        kwargs = ctx.kwargs

        view = ctx.view
        iterator = iter(self.params.items())

        if self.cog is not None:
            try:
                next(iterator)
            except StopIteration:
                fmt = 'Callback for {0.name} command is missing "self" parameter.'
                raise ClientException(fmt.format(self))

        try:
            next(iterator)
        except StopIteration:
            fmt = 'Callback for {0.name} command is missing "ctx" parameter.'
            raise ClientException(fmt.format(self))

        for name, param in iterator:
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                transformed = await self.transform(ctx, param)
                args.append(transformed)
            elif param.kind == param.KEYWORD_ONLY:
                if self.rest_is_raw:
                    converter = self._get_converter(param)
                    argument = view.read_rest()
                    kwargs[name] = await self.do_conversion(ctx, converter, argument, param)
                else:
                    kwargs[name] = await self.transform(ctx, param)
                break
            elif param.kind == param.VAR_POSITIONAL:
                while not view.eof:
                    try:
                        transformed = await self.transform(ctx, param)
                        args.append(transformed)
                    except RuntimeError:
                        break

        if not self.ignore_extra:
            if not view.eof:
                raise TooManyArguments('Too many arguments passed to ' + self.qualified_name)

    async def _verify_checks(self, ctx):
        if not self.enabled:
            raise DisabledCommand('{0.name} command is disabled'.format(self))

        if not await self.can_run(ctx):
            raise CheckFailure('The check functions for command {0.qualified_name} failed.'.format(self))

    async def call_before_hooks(self, ctx):
        cog = self.cog
        if self._before_invoke is not None:
            if cog is None:
                await self._before_invoke(ctx)
            else:
                await self._before_invoke(cog, ctx)

        if cog is not None:
            hook = Cog._get_overridden_method(cog.cog_before_invoke)
            if hook is not None:
                await hook(ctx)

        hook = ctx.bot._before_invoke
        if hook is not None:
            await hook(ctx)

    async def call_after_hooks(self, ctx):
        cog = self.cog
        if self._after_invoke is not None:
            if cog is None:
                await self._after_invoke(ctx)
            else:
                await self._after_invoke(cog, ctx)

        if cog is not None:
            hook = Cog._get_overridden_method(cog.cog_after_invoke)
            if hook is not None:
                await hook(ctx)

        hook = ctx.bot._after_invoke
        if hook is not None:
            await hook(ctx)

    def _prepare_cooldowns(self, ctx):
        if self._buckets.valid:
            current = ctx.message.date.replace(tzinfo=datetime.timezone.utc).timestamp()
            bucket = self._buckets.get_bucket(ctx.message, current)
            retry_after = bucket.update_rate_limit(current)
            if retry_after:
                raise CommandOnCooldown(bucket, retry_after)

    async def prepare(self, ctx):
        ctx.command = self
        await self._verify_checks(ctx)

        if self.cooldown_after_parsing:
            await self._parse_arguments(ctx)
            self._prepare_cooldowns(ctx)
        else:
            self._prepare_cooldowns(ctx)
            await self._parse_arguments(ctx)

        await self.call_before_hooks(ctx)

    def is_on_cooldown(self, ctx):
        """Checks whether the command is currently on cooldown.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to use when checking the commands cooldown status.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the command is on cooldown.
        """
        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(ctx.message)
        return bucket.get_tokens() == 0

    def reset_cooldown(self, ctx):
        """Resets the cooldown on this command.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to reset the cooldown under.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(ctx.message)
            bucket.reset()

    async def invoke(self, ctx):
        await self.prepare(ctx)
        ctx.invoked_subcommand = None
        injected = hooked_wrapped_callback(self, ctx, self.callback)
        await injected(*ctx.args, **ctx.kwargs)

    async def reinvoke(self, ctx, *, call_hooks=False):
        ctx.command = self
        await self._parse_arguments(ctx)

        if call_hooks:
            await self.call_before_hooks(ctx)

        ctx.invoked_subcommand = None
        try:
            await self.callback(*ctx.args, **ctx.kwargs)
        except:
            ctx.command_failed = True
            raise
        finally:
            if call_hooks:
                await self.call_after_hooks(ctx)

    def error(self, coro):
        """A decorator that registers a coroutine as a local error handler.

        A local error handler is an :func:`.on_command_error` event limited to
        a single command. However, the :func:`.on_command_error` is still
        invoked afterwards as the catch-all.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the local error handler.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The error handler must be a coroutine.')

        self.on_error = coro
        return coro

    def before_invoke(self, coro):
        """A decorator that registers a coroutine as a pre-invoke hook.

        A pre-invoke hook is called directly before the command is
        called. This makes it a useful function to set up database
        connections or any type of set up required.

        This pre-invoke hook takes a sole parameter, a :class:`.Context`.

        See :meth:`.Bot.before_invoke` for more info.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the pre-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The pre-invoke hook must be a coroutine.')

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro):
        """A decorator that registers a coroutine as a post-invoke hook.

        A post-invoke hook is called directly after the command is
        called. This makes it a useful function to clean-up database
        connections or any type of clean up required.

        This post-invoke hook takes a sole parameter, a :class:`.Context`.

        See :meth:`.Bot.after_invoke` for more info.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the post-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The post-invoke hook must be a coroutine.')

        self._after_invoke = coro
        return coro

    @property
    def cog_name(self):
        """:class:`str`: The name of the cog this command belongs to. None otherwise."""
        return type(self.cog).__cog_name__ if self.cog is not None else None

    def _is_typing_optional(self, annotation):
        try:
            origin = annotation.__origin__
        except AttributeError:
            return False

        if origin is not typing.Union:
            return False

        return annotation.__args__[-1] is type(None)

    async def can_run(self, ctx):
        """|coro|

        Checks if the command can be executed by checking all the predicates
        inside the :attr:`.checks` attribute. This also checks whether the
        command is disabled.

        Parameters
        -----------
        ctx: :class:`.Context`
            The ctx of the command currently being invoked.

        Raises
        -------
        :class:`CommandError`
            Any command error that was raised during a check call will be propagated
            by this function.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the command can be invoked.
        """

        if not self.enabled:
            raise DisabledCommand('{0.name} command is disabled'.format(self))

        original = ctx.command
        ctx.command = self

        try:
            if not await ctx.bot.can_run(ctx):
                raise CheckFailure('The global check functions for command {0.qualified_name} failed.'.format(self))

            cog = self.cog
            if cog is not None:
                local_check = Cog._get_overridden_method(cog.cog_check)
                if local_check is not None:
                    ret = await maybe_coroutine(local_check, ctx)
                    if not ret:
                        return False

            predicates = self.checks
            if not predicates:
                return True

            return await async_all(predicate(ctx) for predicate in predicates)
        finally:
            ctx.command = original


def cooldown(rate, per, type=BucketType.default):
    """A decorator that adds a cooldown to a :class:`.Command`
    or its subclasses.

    A cooldown allows a command to only be used a specific amount
    of times in a specific time frame. These cooldowns can be based
    either on a per-user, per-conversation, per-member or global basis.

    Denoted by the third argument of ``type`` which must be of enum
    type ``BucketType`` which could be either:

    - ``BucketType.default`` for a global basis.
    - ``BucketType.user`` for a per-user basis.
    - ``BucketType.conversation`` for a per-conversation basis.
    - ``BucketType.member`` for a per-member.

    If a cooldown is triggered, then :exc:`.CommandOnCooldown` is triggered in
    :func:`.on_command_error` and the local error handler.

    A command can only have a single cooldown.

    Parameters
    ------------
    rate: :class:`int`
        The number of times a command can be used before triggering a cooldown.
    per: :class:`float`
        The amount of seconds to wait for a cooldown when it's been triggered.
    type: ``BucketType``
        The type of cooldown to have.
    """

    def decorator(func):
        if isinstance(func, Command):
            func._buckets = CooldownMapping(Cooldown(rate, per, type))
        else:
            func.__commands_cooldown__ = Cooldown(rate, per, type)
        return func

    return decorator



from copy import deepcopy


class Group:
    """Represents a TG Group

    Attributes
    ----------
    id: :class:`int`
        Id of group, positive int
    name: :class:`str`
        Display name of the group
    screen_name: :class:`str`
        Screen name of the group (link part after vk.com/)
    is_closed: :class:`bool`
        ``True`` if group is closed
    type: :class:`str`
        Can be either `event`, `group` or `page` depending on group type
    photo: :class:`dict`
        Has group photo urls with different sizes
    """

    def __init__(self, data):
        self.original_data = deepcopy(data)
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.is_bot = data.get('is_bot')
        self.first_name = data.get('first_name')
        self.last_name = data.get('last_name') if 'last_name' in data else None
        self.username = data.get('username') if 'username' in data else None
        self.language_code = data.get('language_code') if 'language_code' in data else None
        self.is_premium = data.get('is_premium') if 'is_premium' in data else False
        self.added_to_attachment_menu = data.get('added_to_attachment_menu') if 'added_to_attachment_menu' in data else False
        self.can_join_groups = data.get('can_join_groups') if 'can_join_groups' in data else False
        self.can_read_all_group_messages = data.get('can_read_all_group_messages') if 'can_read_all_group_messages' in data else False
        self.supports_inline_queries = data.get('supports_inline_queries') if 'supports_inline_queries' in data else False

class Comments:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.count = data.get('count')
        self.can_post = data.get('can_post')
        self.groups_can_post = data.get('groups_can_post')
        self.can_close = data.get('can_close')
        self.can_open = data.get('can_open')


class Likes:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.count = data.get('count')
        self.user_likes = data.get('user_likes')
        self.can_like = data.get('can_like')
        self.can_publish = data.get('can_publish')


class Reposts:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.count = data.get('count')
        self.user_reposted = data.get('user_reposted')


class Views:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.count = data.get('count')


class Geo:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.type = data.get('type')
        self.coordinates = data.get('coordinates')
        self.place = data.get('place')


class Thread:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.count = data.get('count')
        self.items = data.get('items')
        self.can_post = data.get('can_post')
        self.show_reply_button = data.get('show_reply_button')
        self.groups_can_post = data.get('groups_can_post')


class WallComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_id = data.get('from_id')
        self.date = data.get('date')
        self.text = data.get('text')
        self.reply_to_user = data.get('reply_to_user')
        self.reply_to_comment = data.get('reply_to_comment')
        self.attachments = data.get('attachments')
        self.parents_stack = data.get('parents_stack')
        self.thread = Thread(data.get('thread', {}))
        self.post_id = data.get('post_id')
        self.post_owner_id = data.get('post_owner_id')


class DeletedWallComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.owner_id = data.get('owner_id')
        self.id = data.get('id')
        self.deleter_id = data.get('deleter_id')
        self.post_id = data.get('post_id')


class MarketComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_id = data.get('from_id')
        self.date = data.get('date')
        self.text = data.get('text')
        self.reply_to_user = data.get('reply_to_user')
        self.reply_to_comment = data.get('reply_to_comment')
        self.attachments = data.get('attachments')
        self.parents_stack = data.get('parents_stack')
        self.thread = Thread(data.get('thread', {}))
        self.market_owner_id = data.get('market_owner_id')
        self.item_id = data.get('item_id')


class DeletedMarketComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.owner_id = data.get('owner_id')
        self.id = data.get('id')
        self.user_id = data.get('user_id')
        self.deleter_id = data.get('deleter_id')
        self.item_id = data.get('item_id')


class VideoComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_id = data.get('from_id')
        self.date = data.get('date')
        self.text = data.get('text')
        self.reply_to_user = data.get('reply_to_user')
        self.reply_to_comment = data.get('reply_to_comment')
        self.attachments = data.get('attachments')
        self.parents_stack = data.get('parents_stack')
        self.thread = Thread(data.get('thread', {}))
        self.video_id = data.get('video_id')
        self.video_owner_id = data.get('video_owner_id')


class DeletedVideoComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.owner_id = data.get('owner_id')
        self.id = data.get('id')
        self.user_id = data.get('user_id')
        self.deleter_id = data.get('deleter_id')
        self.video_id = data.get('video_id')


class PhotoComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_id = data.get('from_id')
        self.date = data.get('date')
        self.text = data.get('text')
        self.reply_to_user = data.get('reply_to_user')
        self.reply_to_comment = data.get('reply_to_comment')
        self.attachments = data.get('attachments')
        self.parents_stack = data.get('parents_stack')
        self.thread = Thread(data.get('thread', {}))
        self.photo_id = data.get('photo_id')
        self.photo_owner_id = data.get('photo_owner_id')


class DeletedPhotoComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.owner_id = data.get('owner_id')
        self.id = data.get('id')
        self.user_id = data.get('user_id')
        self.deleter_id = data.get('deleter_id')
        self.photo_id = data.get('photo_id')


class Post:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_id = data.get('from_id')
        self.owner_id = data.get('owner_id')
        self.date = data.get('date')
        self.marked_as_ads = data.get('marked_as_ads')
        self.post_type = data.get('post_type')
        self.text = data.get('text')
        self.can_pin = data.get('can_pin')
        self.can_edit = data.get('can_edit')
        self.created_by = data.get('created_by')
        self.can_delete = data.get('can_delete')
        self.comments = Comments(data.get('comments', {}))
        self.is_favorite = data.get('is_favorite')
        self.likes = Likes(data.get('likes', {}))
        self.reposts = Reposts(data.get('reposts', {}))
        self.views = Views(data.get('views', {}))
        self.attachments = data.get('attachments')
        self.geo = Geo(data.get('geo', {}))
        self.signer_id = data.get('signer_id')
        self.copy_history = data.get('copy_history')
        self.is_pinned = data.get('is_pinned')
        self.postponed_id = data.get('postponed_id')


class BoardComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.id = data.get('id')
        self.from_id = data.get('from_id')
        self.date = data.get('date')
        self.text = data.get('text')
        self.attachments = data.get('attachments')
        self.likes = Likes(data.get('likes', {}))
        self.topic_id = data.get('topic_id')
        self.topic_owner_id = data.get('topic_owner_id')


class DeletedBoardComment:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.topic_owner_id = data.get('topic_owner_id')
        self.topic_id = data.get('topic_id')
        self.id = data.get('id')


class PollVote:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.owner_id = data.get('owner_id')
        self.poll_id = data.get('poll_id')
        self.option_id = data.get('option_id')
        self.user_id = data.get('user_id')


class OfficersEdit:

    def __init__(self, data):
        self._unpack(data)

    def _unpack(self, data):
        self.admin_id = data.get('admin_id')
        self.user_id = data.get('user_id')
        self.level_old = data.get('level_old')
        self.level_new = data.get('level_new')



import asyncio
import collections
import importlib
import inspect
import re
import sys
import traceback
import types

from tg_botting.client import Client
from tg_botting.cog import Cog
from tg_botting.commands import GroupMixin
from tg_botting.context import Context
from tg_botting.exceptions import ExtensionFailed, NoEntryPointError, ExtensionAlreadyLoaded, ExtensionNotFound, \
    ExtensionNotLoaded, CommandError, CommandNotFound
from tg_botting.utils import async_all, maybe_coroutine, find
from tg_botting.view import StringView


def when_mentioned(bot, msg):
    r"""A callable that implements a command prefix equivalent to being mentioned.
    These are meant to be passed into the :attr:`.Bot.command_prefix` attribute.
    """
    match = re.match(r'\[club{}\|[^]]+],? '.format(bot.group.id), msg.text)
    if match and msg.text.startswith(match.group()):
        return [match.group()]
    return ['[club{0.group.id}|@{0.group.screen_name}] '.format(bot), '[club{0.group.id}|{0.group.name}] '.format(bot)]


def when_mentioned_or(*prefixes):
    """A callable that implements when mentioned or other prefixes provided.
    These are meant to be passed into the :attr:`.Bot.command_prefix` attribute.

    Example
    --------

    .. code-block:: python3

        bot = Bot(command_prefix=commands.when_mentioned_or('!'))


    .. note::
        This callable returns another callable, so if this is done inside a custom
        callable, you must call the returned callable, for example:

        .. code-block:: python3

            async def get_prefix(bot, message):
                extras = await prefixes_for(message.peer_id) # returns a list
                return commands.when_mentioned_or(*extras)(bot, message)


    See Also
    ----------
    :func:`.when_mentioned`
    :func:`.when_mentioned_or_pm`
    :func:`.when_mentioned_or_pm_or`
    """

    def inner(bot, msg):
        r = list(prefixes)
        r = when_mentioned(bot, msg) + r
        return r

    return inner


def when_mentioned_or_pm():
    """A callable that implements when mentioned or bot gets pm.
    These are meant to be passed into the :attr:`.Bot.command_prefix` attribute.

    Example
    --------

    .. code-block:: python3

        bot = commands.Bot(command_prefix=commands.when_mentioned_or_pm())


    See Also
    ----------
    :func:`.when_mentioned`
    :func:`.when_mentioned_or`
    :func:`.when_mentioned_or_pm_or`
    """

    def inner(bot, msg):
        r = when_mentioned(bot, msg)
        if msg.peer_id == msg.from_id:
            r.append('')
        return r

    return inner


def when_mentioned_or_pm_or(*prefixes):
    """A callable that implements when mentioned, pm or other prefixes provided.
    These are meant to be passed into the :attr:`.Bot.command_prefix` attribute.

    Example
    --------

    .. code-block:: python3

        bot = commands.Bot(command_prefix=commands.when_mentioned_or_pm_or('!'))

    .. note::

        This callable returns another callable, so if this is done inside a custom
        callable, you must call the returned callable, for example:

        .. code-block:: python3

            async def get_prefix(bot, message):
                extras = await prefixes_for(message.peer_id) # returns a list
                return commands.when_mentioned_or(*extras)(bot, message)


    See Also
    ----------
    :func:`.when_mentioned`
    :func:`.when_mentioned_or`
    :func:`.when_mentioned_or_pm`
    """

    def inner(bot, msg):
        r = when_mentioned(bot, msg) + list(prefixes)
        if msg.peer_id == msg.from_id:
            r.append('')
        return r

    return inner


def _is_submodule(parent, child):
    return parent == child or child.startswith(parent + ".")


class BotBase(GroupMixin):
    def __init__(self, command_prefix, description=None, **options):
        super().__init__(**options)
        self.command_prefix = command_prefix
        self.extra_events = {}
        self.__cogs = {}
        self.__extensions = {}
        self._checks = []
        self._check_once = []
        self._before_invoke = None
        self._after_invoke = None
        self.description = inspect.cleandoc(description) if description else ''
        self.owner_id = options.get('owner_id')

        if options.pop('self_bot', False):
            self._skip_check = lambda x, y: x != y
        else:
            self._skip_check = lambda x, y: x == y

    def dispatch(self, event_name, *args, **kwargs):
        super().dispatch(event_name, *args, **kwargs)
        ev = 'on_' + event_name
        for event in self.extra_events.get(ev, []):
            self._schedule_event(event, ev, *args, **kwargs)

    async def close(self):
        for extension in tuple(self.__extensions):
            try:
                self.unload_extension(extension)
            except Exception:
                pass

        for cog in tuple(self.__cogs):
            try:
                self.remove_cog(cog)
            except Exception:
                pass

        await super().close()

    async def on_command_error(self, context, exception):
        if self.extra_events.get('on_command_error', None):
            return

        if hasattr(context.command, 'on_error'):
            return

        cog = context.cog
        if cog:
            if Cog._get_overridden_method(cog.cog_command_error) is not None:
                return

        print('Ignoring exception in command {}:'.format(context.command), file=sys.stderr)
        traceback.print_exception(type(exception), exception, exception.__traceback__, file=sys.stderr)

    def check(self, func):
        self.add_check(func)
        return func

    def add_check(self, func, *, call_once=False):
        if call_once:
            self._check_once.append(func)
        else:
            self._checks.append(func)

    def remove_check(self, func, *, call_once=False):
        l = self._check_once if call_once else self._checks

        try:
            l.remove(func)
        except ValueError:
            pass

    def check_once(self, func):
        self.add_check(func, call_once=True)
        return func

    async def can_run(self, ctx, *, call_once=False):
        data = self._check_once if call_once else self._checks

        if len(data) == 0:
            return True

        return await async_all(f(ctx) for f in data)

    def before_invoke(self, coro):
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The pre-invoke hook must be a coroutine.')

        self._before_invoke = coro
        return coro

    def after_invoke(self, coro):
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The post-invoke hook must be a coroutine.')

        self._after_invoke = coro
        return coro

    def add_listener(self, func, name=None):
        name = func.__name__ if name is None else name

        if not asyncio.iscoroutinefunction(func):
            raise TypeError('Listeners must be coroutines')

        if name in self.extra_events:
            self.extra_events[name].append(func)
        else:
            self.extra_events[name] = [func]

    def remove_listener(self, func, name=None):
        name = func.__name__ if name is None else name

        if name in self.extra_events:
            try:
                self.extra_events[name].remove(func)
            except ValueError:
                pass

    def listen(self, name=None):
        def decorator(func):
            self.add_listener(func, name)
            return func

        return decorator

    def add_cog(self, cog):
        if not isinstance(cog, Cog):
            raise TypeError('cogs must derive from Cog')

        cog = cog._inject(self)
        self.__cogs[cog.__cog_name__] = cog

    def get_cog(self, name):
        return self.__cogs.get(name)

    def remove_cog(self, name):
        cog = self.__cogs.pop(name, None)
        if cog is None:
            return
        cog._eject(self)

    @property
    def cogs(self):
        return types.MappingProxyType(self.__cogs)

    def _remove_module_references(self, name):
        for cogname, cog in self.__cogs.copy().items():
            if _is_submodule(name, cog.__module__):
                self.remove_cog(cogname)

        for cmd in self.all_commands.copy().values():
            if cmd.module is not None and _is_submodule(name, cmd.module):
                if isinstance(cmd, GroupMixin):
                    cmd.recursively_remove_all_commands()
                self.remove_command(cmd.name)

        for event_list in self.extra_events.copy().values():
            remove = []
            for index, event in enumerate(event_list):
                if event.__module__ is not None and _is_submodule(name, event.__module__):
                    remove.append(index)

            for index in reversed(remove):
                del event_list[index]

    def _call_module_finalizers(self, lib, key):
        try:
            func = getattr(lib, 'teardown')
        except AttributeError:
            pass
        else:
            try:
                func(self)
            except Exception:
                pass
        finally:
            self.__extensions.pop(key, None)
            sys.modules.pop(key, None)
            name = lib.__name__
            for module in list(sys.modules.keys()):
                if _is_submodule(name, module):
                    del sys.modules[module]

    def _load_from_module_spec(self, lib, key):
        try:
            setup = getattr(lib, 'setup')
        except AttributeError:
            del sys.modules[key]
            raise NoEntryPointError(key)

        try:
            setup(self)
        except Exception as e:
            self._remove_module_references(lib.__name__)
            self._call_module_finalizers(lib, key)
            raise ExtensionFailed(key, e) from e
        else:
            self.__extensions[key] = lib

    def load_extension(self, name):
        """Loads an extension.
        An extension is a python module that contains commands, cogs, or
        listeners.
        An extension must have a global function, ``setup`` defined as
        the entry point on what to do when the extension is loaded. This entry
        point must have a single argument, the ``bot``.

        Parameters
        ------------
        name: :class:`str`
            The extension name to load. It must be dot separated like
            regular Python imports if accessing a sub-module. e.g.
            ``foo.test`` if you want to import ``foo/test.py``.

        Raises
        --------
        ExtensionNotFound
            The extension could not be imported.
        ExtensionAlreadyLoaded
            The extension is already loaded.
        NoEntryPointError
            The extension does not have a setup function.
        ExtensionFailed
            The extension or its setup function had an execution error.
        """
        if name in self.__extensions:
            raise ExtensionAlreadyLoaded(name)

        try:
            lib = importlib.import_module(name)
        except ImportError as e:
            raise ExtensionNotFound(name, e) from e
        else:
            self._load_from_module_spec(lib, name)

    def unload_extension(self, name):
        """Unloads an extension.
        When the extension is unloaded, all commands, listeners, and cogs are
        removed from the bot and the module is un-imported.
        The extension can provide an optional global function, ``teardown``,
        to do miscellaneous clean-up if necessary. This function takes a single
        parameter, the ``bot``, similar to ``setup`` from
        :meth:`~.Bot.load_extension`.

        Parameters
        ------------
        name: :class:`str`
            The extension name to unload. It must be dot separated like
            regular Python imports if accessing a sub-module. e.g.
            ``foo.test`` if you want to import ``foo/test.py``.

        Raises
        -------
        ExtensionNotLoaded
            The extension was not loaded.
        """
        lib = self.__extensions.get(name)
        if lib is None:
            raise ExtensionNotLoaded(name)

        self._remove_module_references(lib.__name__)
        self._call_module_finalizers(lib, name)

    def reload_extension(self, name):
        """Atomically reloads an extension.
        This replaces the extension with the same extension, only refreshed. This is
        equivalent to a :meth:`unload_extension` followed by a :meth:`load_extension`
        except done in an atomic way. That is, if an operation fails mid-reload then
        the bot will roll-back to the prior working state.

        Parameters
        ------------
        name: :class:`str`
            The extension name to reload. It must be dot separated like
            regular Python imports if accessing a sub-module. e.g.
            ``foo.test`` if you want to import ``foo/test.py``.

        Raises
        -------
        ExtensionNotLoaded
            The extension was not loaded.
        ExtensionNotFound
            The extension could not be imported.
        NoEntryPointError
            The extension does not have a setup function.
        ExtensionFailed
            The extension setup function had an execution error.
        """
        lib = self.__extensions.get(name)
        if lib is None:
            raise ExtensionNotLoaded(name)

        modules = {
            name: module
            for name, module in sys.modules.items()
            if _is_submodule(lib.__name__, name)
        }

        try:
            self._remove_module_references(lib.__name__)
            self._call_module_finalizers(lib, name)
            self.load_extension(name)
        except Exception as e:
            self._load_from_module_spec(lib, name)
            sys.modules.update(modules)
            raise

    @property
    def extensions(self):
        """Mapping[:class:`str`, :class:`py:types.ModuleType`]: A read-only mapping of extension name to extension."""
        return types.MappingProxyType(self.__extensions)

    async def get_prefix(self, message):
        """|coro|

        Retrieves the prefix the bot is listening to
        with the message as a context.

        Parameters
        -----------
        message: :class:`.Message`
            The message context to get the prefix of.

        Returns
        --------
        Union[List[:class:`str`], :class:`str`]
            A list of prefixes or a single prefix that the bot is
            listening for.
        """
        prefix = ret = self.command_prefix
        if callable(prefix):
            ret = await maybe_coroutine(prefix, self, message)

        if not isinstance(ret, str):
            try:
                ret = list(ret)
            except TypeError:
                if isinstance(ret, collections.Iterable):
                    raise

                raise TypeError("command_prefix must be plain string, iterable of strings, or callable "
                                "returning either of these, not {}".format(ret.__class__.__name__))

            if not ret:
                raise ValueError("Iterable command_prefix must contain at least one prefix")

        return ret

    async def get_context(self, message, *, cls=Context):
        view = StringView(message.text)
        ctx = cls(prefix=None, view=view, bot=self, message=message)

        if self._skip_check(message.chat.id, self.group.id):
            return ctx

        prefix = await self.get_prefix(message)
        invoked_prefix = prefix

        if isinstance(prefix, str):
            if not view.skip_string(prefix):
                return ctx
        else:
            try:
                if message.text.startswith(tuple(prefix)):
                    invoked_prefix = find(view.skip_string, prefix)
                else:
                    return ctx

            except TypeError:
                if not isinstance(prefix, list):
                    raise TypeError("get_prefix must return either a string or a list of string, "
                                    "not {}".format(prefix.__class__.__name__))

                for value in prefix:
                    if not isinstance(value, str):
                        raise TypeError("Iterable command_prefix or list returned from get_prefix must "
                                        "contain only strings, not {}".format(value.__class__.__name__))

                raise

        msg = view.read_rest()
        view.undo()
        # This is by far the fastest method there is as far as my knowledge extends
        # Considering VK message length limit is 4096 it is at most 2048 iterations which is at most about 50 ms
        # If not exaggerating the length of user messages this is the most optimal way imo
        # It also works well with several commands having the same beginning as it will always choose the longest one
        commands = set(self.all_commands) if not self.case_insensitive else self.all_commands
        words = msg.split(' ')
        for wordamt in range(len(words), 1, -1):
            potcomm = ' '.join(words[:wordamt])
            if potcomm in commands:
                invoker = potcomm
                view.read(len(invoker))
                break
        else:
            invoker = view.get_word()
        ctx.invoked_with = invoker
        ctx.prefix = invoked_prefix
        ctx.command = self.all_commands.get(invoker)
        return ctx

    async def invoke(self, ctx):
        """|coro|

        Invokes the command given under the invocation context and
        handles all the internal event dispatch mechanisms.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to invoke.
        """
        if ctx.command is not None:
            self.dispatch('command', ctx)
            try:
                if await self.can_run(ctx, call_once=True):
                    await ctx.command.invoke(ctx)
            except CommandError as exc:
                await ctx.command.dispatch_error(ctx, exc)
            else:
                self.dispatch('command_completion', ctx)
        elif ctx.invoked_with:
            exc = CommandNotFound('Command "{}" is not found'.format(ctx.invoked_with))
            self.dispatch('command_error', ctx, exc)

    async def process_commands(self, message):
        if message.chat.id == self.group.id:
            return

        ctx = await self.get_context(message)
        await self.invoke(ctx)

    async def on_message_new(self, message):
        await self.process_commands(message)


class Bot(BotBase, Client):
    pass



_title__ = 'tg_botting'
__author__ = 'Sweetie'
__license__ = 'MIT'
__copyright__ = 'Copyright 2023-present Sweetie'
__version__ = '2.1.4'

from collections import namedtuple
import logging

from .bot import Bot
from .limiters import *
from .conversions import Converter
from .message import *
from .user import *
from .utils import *
from .permissions import *
from .objects import *
from .commands import *


VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')

version_info = VersionInfo(major=0, minor=11, micro=0, releaselevel='development', serial=0)

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

logging.getLogger(__name__).addHandler(NullHandler())



from tg_botting.abstract import Messageable


class Context(Messageable):
    def __init__(self, **attrs):
        self.message = attrs.pop('message', None)
        self.bot = attrs.pop('bot', None)
        self.args = attrs.pop('args', [])
        self.kwargs = attrs.pop('kwargs', {})
        self.prefix = attrs.pop('prefix')
        self.command = attrs.pop('command', None)
        self.view = attrs.pop('view', None)
        self.invoked_with = attrs.pop('invoked_with', None)
        self.invoked_subcommand = attrs.pop('invoked_subcommand', None)
        self.subcommand_passed = attrs.pop('subcommand_passed', None)
        self.command_failed = attrs.pop('command_failed', False)

    async def invoke(self, *args, **kwargs):
        try:
            command = args[0]
        except IndexError:
            raise TypeError('Missing command to invoke.') from None

        arguments = []
        if command.cog is not None:
            arguments.append(command.cog)

        arguments.append(self)
        arguments.extend(args[1:])

        ret = await command.callback(*arguments, **kwargs)
        return ret

    async def reinvoke(self, *, call_hooks=False, restart=True):
        cmd = self.command
        view = self.view
        if cmd is None:
            raise ValueError('This context is not valid.')

        index, previous = view.index, view.previous
        invoked_with = self.invoked_with
        invoked_subcommand = self.invoked_subcommand
        subcommand_passed = self.subcommand_passed

        if restart:
            to_call = cmd.root_parent or cmd
            view.index = len(self.prefix)
            view.previous = 0
            view.get_word()
        else:
            to_call = cmd

        try:
            await to_call.reinvoke(self, call_hooks=call_hooks)
        finally:
            self.command = cmd
            view.index = index
            view.previous = previous
            self.invoked_with = invoked_with
            self.invoked_subcommand = invoked_subcommand
            self.subcommand_passed = subcommand_passed

    async def reply(self, text, **kwargs):
        return await self.send(text, reply_to_message_id=self.message.message_id,**kwargs)

    @property
    def cog(self):
        """Returns the cog associated with this context's command. None if it does not exist."""
        if self.command is None:
            return None
        return self.command.cog

    @property
    def valid(self):
        """Checks if the invocation context is valid to be invoked with."""
        return self.prefix is not None and self.command is not None

    @property
    def author(self):
        """Shorthand for :attr:`.Message.from_id`"""
        return self.message.user

    @property
    def from_id(self):
        """Shorthand for :attr:`.Message.from_id`"""
        return self.message.user.id

    @property
    def peer_id(self):
        """Shorthand for :attr:`.Message.peer_id`"""
        return self.message.chat.id

    @property
    def user(self):
        return self.message.user

    @property
    def chat(self):
        return self.message.chat

    @property
    def text(self):
        """Shorthand for :attr:`.Message.text`"""
        return self.message.text

    async def _get_conversation(self):
        return self.message.chat.id

    @property
    def me(self):
        """Returns bot :class:`.Group` or :class:`.User`, depending on whether it is :class:`.Bot` or :class:`.UserBot`"""
        return self.bot.group or self.bot.user
